# OTHER IMPORTS
from typing import Optional, List

# ROS IMPORTS
import gymnasium as gym

# EAGERx IMPORTS
from eagerx.core.space import Space
from eagerx.engines.openai_gym.engine import GymEngine
from eagerx.core.entities import Object
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register


class GymObject(Object):
    @classmethod
    @register.sensors(
        observation=Space(dtype="float32"),
        reward=Space(dtype="float32"),
        terminated=Space(low=0, high=1, shape=(), dtype="int64"),
        truncated=Space(low=0, high=1, shape=(), dtype="int64"),
        image=Space(dtype="uint8"),
    )
    @register.actuators(action=Space(dtype="float32"))
    @register.engine_states()
    def make(
        cls,
        name: str,
        sensors: Optional[List[str]] = None,
        env_id: str = "Pendulum-v1",
        rate: float = 20.0,
        always_render: bool = False,
        default_action=None,
        render_shape: Optional[List[int]] = None,
    ):
        """Object spec of GymObject"""
        spec = cls.get_specification()

        # Set default node params
        spec.config.name = name
        spec.config.sensors = sensors if isinstance(sensors, list) else ["observation", "reward", "truncated", "terminated"]
        spec.config.actuators = ["action"]

        # Set custom node params
        spec.config.render_shape = render_shape if render_shape else [200, 200]
        spec.config.env_id = env_id
        spec.config.always_render = always_render
        spec.config.default_action = default_action

        # Set spaces
        env = gym.make(spec.config.env_id)
        obs_space = env.observation_space
        if isinstance(obs_space, gym.spaces.Discrete):
            spec.sensors.observation.space = Space(low=0, high=obs_space.n - 1, shape=(), dtype="int64")
        elif isinstance(obs_space, gym.spaces.Box):
            spec.sensors.observation.space = Space(obs_space.low, obs_space.high, shape=obs_space.shape, dtype=obs_space.dtype)
        else:
            raise NotImplementedError("Space not compatible with this object.")
        act_space = env.action_space
        if isinstance(act_space, gym.spaces.Discrete):
            spec.actuators.action.space = Space(low=0, high=act_space.n - 1, shape=(), dtype="int64")
        elif isinstance(act_space, gym.spaces.Box):
            spec.actuators.action.space = Space(act_space.low, act_space.high, shape=act_space.shape, dtype=act_space.dtype)
        else:
            raise NotImplementedError("Space not compatible with this object.")

        spec.sensors.reward.space = Space(low=env.reward_range[0], high=env.reward_range[1], shape=tuple(), dtype="float32")
        shape = (spec.config.render_shape[0], spec.config.render_shape[1], 3)
        spec.sensors.image.space = Space(low=0, high=255, shape=shape, dtype="uint8")

        # Set rates
        spec.sensors.observation.rate = rate
        spec.sensors.reward.rate = rate
        spec.sensors.truncated.rate = rate
        spec.sensors.terminated.rate = rate
        spec.sensors.image.rate = rate
        spec.actuators.action.rate = rate
        return spec

    @staticmethod
    @register.engine(GymEngine)  # This decorator pre-initializes engine implementation with default object_params
    def openai_gym(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (GymEngine) of the object."""
        # Set engine arguments (nothing to set here in this case)
        spec.engine.env_id = spec.config.env_id

        # Create sensor engine nodes
        # Rate=None, because we will connect them to sensors (thus uses the rate set in the agnostic specification)
        from eagerx.engines.openai_gym.enginenodes import (
            ObservationSensor,
            RewardSensor,
            TruncatedSensor,
            TerminatedSensor,
            ActionActuator,
            GymImage,
        )

        obs = ObservationSensor.make("obs", rate=spec.sensors.observation.rate, process=2)
        obs.outputs.observation.space = spec.sensors.observation.space
        rwd = RewardSensor.make("rwd", rate=spec.sensors.reward.rate, process=2)
        truncated = TruncatedSensor.make("truncated", rate=spec.sensors.truncated.rate, process=2)
        terminated = TerminatedSensor.make("terminated", rate=spec.sensors.terminated.rate, process=2)
        image = GymImage.make(
            "image",
            shape=spec.config.render_shape,
            always_render=spec.config.always_render,
            rate=spec.sensors.image.rate,
            process=2,
        )

        # Create actuator engine nodes
        # Rate=None, because we will connect it to an actuator (thus uses the rate set in the agnostic specification)
        action = ActionActuator.make(
            "action", rate=spec.actuators.action.rate, process=2, zero_action=spec.config.default_action
        )
        action.inputs.action.space = spec.actuators.action.space
        action.outputs.action_applied.space = spec.actuators.action.space

        # Connect all engine nodes
        graph.add([obs, rwd, truncated, terminated, image, action])
        graph.connect(source=obs.outputs.observation, sensor="observation")
        graph.connect(source=rwd.outputs.reward, sensor="reward")
        graph.connect(source=truncated.outputs.truncated, sensor="truncated")
        graph.connect(source=terminated.outputs.terminated, sensor="terminated")
        graph.connect(source=image.outputs.image, sensor="image")
        graph.connect(actuator="action", target=action.inputs.action)

        # Check graph validity (comment out)
        # graph.is_valid(plot=True)



import functools
import copy

import eagerx.core.log as log
from eagerx.utils.utils import deepcopy
from eagerx.core.space import Space
from typing import TYPE_CHECKING, Callable, Any
import os


if TYPE_CHECKING:
    from eagerx.core.graph_engine import EngineGraph  # noqa: F401
    from eagerx.core.Entities import Entity, Engine  # noqa: F401

# Global registry with registered entities (engines, objects, nodes, converters, simnodes, etc..)
REGISTRY = dict()
# Global registry with registered I/O types of (.callback, .reset/.pre_reset, .initialize, .add_object)
TYPE_REGISTER = dict()


class ReverseRegisterLookup:
    def __init__(self, d):
        self._dict = d

    def __getitem__(self, spec_lookup):
        for _entity_cls, entity_id in self._dict.items():
            for entity_idd, entry in entity_id.items():
                if entry["spec"] == spec_lookup:
                    return entity_idd


class LookupType:
    def __init__(self, d):
        self._dict = d

    @deepcopy
    def __getitem__(self, func_lookup):
        name_split = func_lookup.__qualname__.split(".")
        cls_name = name_split[0]
        entity_id = func_lookup.__module__ + "/" + cls_name
        return self._dict[entity_id]


# Global (reversed) registry of REGISTER:
REVERSE_REGISTRY = ReverseRegisterLookup(REGISTRY)
LOOKUP_TYPES = LookupType(TYPE_REGISTER)


# TYPES


def _register_types(TYPE_REGISTER, component, cnames, func, space_only=True):
    name_split = func.__qualname__.split(".")
    cls_name = name_split[0]
    fn_name = name_split[1]
    entity_id = func.__module__ + "/" + cls_name
    entry = func.__module__ + "/" + func.__qualname__
    if space_only:
        for key, space in cnames.items():
            if space is None:
                continue
            flag = isinstance(space, Space)
            assert (
                flag
            ), f'TYPE REGISTRATION ERROR: [{cls_name}][{fn_name}][{component}]: "{space}" is an invalid space. Please provide a valid space for "{key}"instead.'
    log.logdebug(f"[{cls_name}][{fn_name}]: {component}={cnames}, entry={entry}")

    @functools.wraps(func)
    def registered_fn(*args, **kwargs):
        """Call the registered function"""
        return func(*args, **kwargs)

    if entity_id not in TYPE_REGISTER:
        """Add class if this is the first registration of class kind"""
        TYPE_REGISTER[entity_id] = dict()

    if component in TYPE_REGISTER[entity_id]:
        """Check if already registered component of duplicate function matches."""
        log.logdebug(f"[{entity_id}][{component}]: {component}={cnames}, entry={entry}")
        flag = cnames == TYPE_REGISTER[entity_id][component] or bool(eval(os.environ.get("EAGERX_RELOAD", "0")))
        assert (
            flag
        ), f'There is already a [{entity_id}][{component}] registered with cnames "{TYPE_REGISTER[entity_id][component]}", and they do not match the cnames of this function: "{cnames}".'
    TYPE_REGISTER[entity_id][component] = cnames
    return registered_fn


def inputs(**inputs: Any) -> Callable:
    """A decorator to register the inputs to a :func:`~eagerx.core.entities.Node.callback`.

    The :func:`~eagerx.core.entities.Node.callback` method should be decorated.

    :param inputs: The input's msg_type class.
    """
    return functools.partial(_register_types, TYPE_REGISTER, "inputs", inputs)


def outputs(**outputs) -> Callable:
    """A decorator to register the outputs of a :func:`~eagerx.core.entities.Node.callback`.

    The :func:`~eagerx.core.entities.Node.callback` method should be decorated.

    :param outputs: The output's msg_type class.
    """
    return functools.partial(_register_types, TYPE_REGISTER, "outputs", outputs)


def states(**states) -> Callable:
    """A decorator to register the states for a :func:`~eagerx.core.entities.Node.reset`.

    The :func:`~eagerx.core.entities.Node.reset` method should be decorated.

    :param outputs: The state's msg_type class.
    """
    return functools.partial(_register_types, TYPE_REGISTER, "states", states)


def targets(**targets) -> Callable:
    """A decorator to register the targets of a :func:`~eagerx.core.entities.ResetNode.callback`.

    The :func:`~eagerx.core.entities.ResetNode.callback` method should be decorated.

    :param targets: The target's msg_type class.
    """
    return functools.partial(_register_types, TYPE_REGISTER, "targets", targets)


def sensors(**sensors) -> Callable:
    """A decorator to register the sensors of an :class:`~eagerx.core.entities.Object`.

    The :func:`~eagerx.core.entities.Object.agnostic` method should be decorated.

    :param sensors: The sensor's msg_type class.
    """
    return functools.partial(_register_types, TYPE_REGISTER, "sensors", sensors)


def actuators(**actuators) -> Callable:
    """A decorator to register the actuators of an :class:`~eagerx.core.entities.Object`.

    The :func:`~eagerx.core.entities.Object.agnostic` method should be decorated.

    :param actuators: The actuator's msg_type class.
    """
    return functools.partial(_register_types, TYPE_REGISTER, "actuators", actuators)


def engine_states(**engine_states) -> Callable:
    """A decorator to register the engine states of an :class:`~eagerx.core.entities.Object`.

    The :func:`~eagerx.core.entities.Object.agnostic` method should be decorated.

    :param engine_states: The engine state's msg_type class.
    """
    return functools.partial(_register_types, TYPE_REGISTER, "engine_states", engine_states)


# ENGINES


def engine(engine_cls: "Engine", entity=None) -> Callable:
    """A decorator to register an engine implementation of an :class:`~eagerx.core.entities.Object`.

    .. note:: In our running example, the :func:`~eagerx.core.entities.Object.example_engine` method would be decorated.

    :param engine_cls: The Engine's subclass (not the baseclass :class:`~eagerx.core.entities.Engine`).
    :param entity: The entity that corresponds to the engine implementation. If left unspecified, the engine is
                   registered to the class that owns the method.
    """
    from eagerx.utils.utils import get_default_params

    engine_config = get_default_params(engine_cls.add_object)
    assert "name" in engine_config, "The object `name` must be an argument to engine.add_object()"
    engine_id = engine_cls.__module__ + "/" + engine_cls.__qualname__

    def _register(func, engine_id=engine_id):
        name_split = func.__qualname__.split(".")
        if entity is not None:
            cls_name = entity.__qualname__
            fn_name = name_split[0]
            entity_id = entity.__module__ + "/" + cls_name
        elif len(name_split) > 1:
            cls_name = name_split[0]
            fn_name = name_split[1]
            entity_id = func.__module__ + "/" + cls_name
        else:
            cls_name = "N/A"
            fn_name = name_split[0]
            entity_id = func.__module__ + "/" + cls_name
        entry = func.__module__ + "/" + func.__qualname__

        log.logdebug(f"[{cls_name}][{fn_name}]: entry={entry}")

        @functools.wraps(func)
        def _engine(spec, engine) -> "EngineGraph":
            """First, initialize spec with object_info, then call the engine function"""
            # Add default engine_config parameters
            engine._initialize_engine_config(spec, copy.deepcopy(engine_config))
            # Initialize engine graph
            graph = spec._initialize_object_graph()
            # Modify engine_config with user-defined engine implementation
            func(spec, graph)
            # Add engine-specific parameters to engine.objects
            engine_params = {key: value for key, value in spec.engine.items() if key != "states"}
            engine.add_object(**engine_params)
            # Add engine-specific states to engine.objects
            engine._add_engine_states(engine_params["name"], spec)
            return graph

        msg = f"Cannot register engine '{entry}' for object '{entity_id}'. "

        # Check if spec of duplicate entity_id corresponds to same spec function
        if entity_id not in REGISTRY:
            REGISTRY[entity_id] = {}
        flag = engine_id in REGISTRY[entity_id] and _engine == REGISTRY[entity_id][engine_id]
        flag = not flag or bool(eval(os.environ.get("EAGERX_RELOAD", "0")))
        assert flag, msg + "There is already an engine implementation of this type registered."

        # Register engine implementation
        REGISTRY[entity_id][engine_id] = _engine
        return _engine

    return _register


def add_engine(spec, engine):
    """Add engine based on registered entity_id"""
    entity_id = spec.config.entity_id
    engine_id = engine.config.entity_id

    msg = f"Cannot add object '{entity_id}' to engine '{engine_id}'. "
    assert entity_id in REGISTRY, msg + f"The Object '{entity_id}' has not been registered yet."
    msg_2 = (
        "If launching the environment as a subprocess, "
        "make sure to reload (i.e. import) the engine implementations in each subprocess."
    )
    assert engine_id in REGISTRY[entity_id], msg + "No engine implementation was registered. " + msg_2
    graph = REGISTRY[entity_id][engine_id](spec, engine)
    return graph



# OTHER
import time
from typing import List, NamedTuple, Any, Tuple
import importlib
import inspect
from functools import wraps
import copy
import json

from eagerx.core.constants import BackendException, Unspecified


def dict_null(items):
    result = {}
    for key, value in items:
        if value is None:
            value = "null"
        result[key] = value
    return result


def dict_None(items):
    result = {}
    for key, value in items:
        if value == "null":
            value = None
        result[key] = value
    return result


def replace_None(d, to_null=True):
    dict_str = json.dumps(d)
    if to_null:
        object_pairs_hook = dict_null
    else:
        object_pairs_hook = dict_None
    return json.loads(dict_str, object_pairs_hook=object_pairs_hook)


def get_attribute_from_module(attribute, module=None):
    if module is None:
        module, attribute = attribute.split("/")
    module = importlib.import_module(module)
    attribute = getattr(module, attribute)
    return attribute


def load(mod_attr: str):
    return get_attribute_from_module(mod_attr)


def initialize_processor(spec):
    processor = load(spec.params["processor_type"])()
    processor.initialize(spec)
    return processor


def is_compatible(dtype_source, dtype_target):
    msg = f"Dtype of source ({dtype_source}) does not match with the dtype of target ({dtype_target})."
    assert dtype_target == dtype_source, msg


class Header(NamedTuple):
    """A dataclass for meta data of sent messages."""

    #: Sequence number of received message since the last reset.
    seq: int
    #: Timestamp according to the simulated clock (seconds). This time is scaled by the real-time factor if > 0.
    sc: float
    #: Timestamp according to the wall clock (seconds).
    wc: float


class Stamp(NamedTuple):
    """A dataclass for timestamping received messages."""

    #: Sequence number of received message.
    seq: int
    #: Timestamp according to the simulated clock (seconds). This time is scaled by the real-time factor if > 0.
    sc: float
    #: Timestamp according to the wall clock (seconds).
    wc: float


class Info(NamedTuple):
    """A dataclass containing info about the received messages in :attr:`~eagerx.utils.utils.Msg.msgs`."""

    #: Name of the registered input.
    name: str
    #: Number of times :func:`~eagerx.core.entities.Node.callback` has been called since the last reset.
    node_tick: int
    #: Rate (Hz) of the input.
    rate_in: float
    #: Simulated timestamp that states during which cycle the message was received since the last reset according
    #: to :attr:`~eagerx.core.entities.Node.rate` and :attr:`~eagerx.utils.utils.Info.node_tick`.
    t_node: List[Stamp]
    #: Simulated timestamp that states at what time the message was received
    #: according to :attr:`~eagerx.utils.utils.Info.rate_in` and :attr:`~eagerx.utils.utils.Stamp.seq`.
    t_in: List[Stamp]
    #: Only concerns states. False if a state has not yet been reset.
    done: bool


class Msg(NamedTuple):
    """A dataclass representing a (windowed) input that is passed to :func:`~eagerx.core.entities.Node.callback`."""

    #: Info on the received messages in :attr:`~eagerx.utils.utils.Msg.msgs`.
    info: Info
    #: The received messages with indexing `msgs[-1]` being the most recent message and `msgs[0]` the oldest.
    msgs: List[Any]


# Set default values
Header.__new__.__defaults__ = (None,) * len(Header._fields)
Stamp.__new__.__defaults__ = (None,) * len(Stamp._fields)
Info.__new__.__defaults__ = (None,) * len(Info._fields)


def deepcopy(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return copy.deepcopy(func(*args, **kwargs))

    return wrapper


def is_supported_type(param: Any, types: Tuple, none_support):
    if isinstance(param, types) or (param is None and none_support):
        if isinstance(param, dict):
            for key, value in param.items():
                assert isinstance(key, str), f'Invalid key "{key}". Only type "str" is supported as dictionary key.'
                is_supported_type(value, types, none_support)
        elif not isinstance(param, str) and hasattr(param, "__iter__"):
            for value in param:
                is_supported_type(value, types, none_support)
    else:
        raise TypeError(
            f'Type "{type(param)}" of a specified (nested) param "{param}" is not supported. Only types {types} are supported.'
        )


def supported_types(*types: Tuple, is_classmethod=True):
    # Check if we support NoneType
    none_support = False
    for a in types:
        if a is None:
            none_support = True
            break

    # Remove None from types
    types = tuple([t for t in types if t is not None])

    def _check(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if is_classmethod:
                check_args = list(args[1:]) + [value for _, value in kwargs.items()]
            else:
                check_args = list(args) + [value for _, value in kwargs.items()]
            for param in check_args:
                is_supported_type(param, types, none_support)
            return func(*args, **kwargs)

        return wrapper

    return _check


def get_default_params(func):
    argspec = inspect.getfullargspec(func)
    if argspec.defaults:
        positional_count = len(argspec.args) - len(argspec.defaults)
        defaults = dict(zip(argspec.args[positional_count:], argspec.defaults))
    else:
        defaults = dict()
        positional_count = len(argspec.args)
    for arg in argspec.args[:positional_count]:
        if arg == "self":
            continue
        defaults[arg] = None
    return defaults


# A singleton that is used to check if an argument was specified.
_unspecified = Unspecified()


def get_param_with_blocking(name, backend, default=_unspecified, timeout=2.0):
    params = Unspecified()
    start = time.time()
    while isinstance(params, Unspecified):
        if time.time() - start > timeout:
            raise KeyError(f"Timeout. Parameter '{name}' not available on parameter server.")
        try:
            params = backend.get_param(name, default=default)
        except (BackendException, KeyError):
            if not isinstance(default, Unspecified):
                return default
            sleep_time = 0.01
            time.sleep(sleep_time)
    return replace_None(params, to_null=False)



# OTHER
from typing import Optional
import gymnasium as gym

# RX IMPORTS
from eagerx.core.constants import process, ERROR
import eagerx.core.register as register
from eagerx.core.entities import Engine


class GymEngine(Engine):
    @classmethod
    def make(
        cls,
        rate,
        process: Optional[int] = process.NEW_PROCESS,
        sync: Optional[bool] = True,
        real_time_factor: Optional[float] = 0,
        simulate_delays: Optional[bool] = True,
        log_level: Optional[int] = ERROR,
    ):
        """TestEngine spec"""
        spec = cls.get_specification()

        # Modify default engine params
        spec.config.rate = rate
        spec.config.process = process
        spec.config.sync = sync
        spec.config.real_time_factor = real_time_factor
        spec.config.simulate_delays = simulate_delays
        spec.config.log_level = log_level
        spec.config.color = "magenta"
        return spec

    def initialize(self, spec):
        self.simulator = dict()

    def add_object(self, name: str, env_id: str, seed: int = None):
        """
        Adds an object whose dynamics are governed by a registered OpenAI gym environment.

        :param name: Name of the :class:`~eagerx.core.entities.Object` that is to be added.
        :param env_id: Gym id of the environment.
        :param seed: Seed for the environment.
        """
        # add object to simulator (we have a ref to the simulator with self.simulator)
        self.backend.loginfo(f'Adding object "{name}" to the simulator.')

        # Create new env, and add to simulator
        env = gym.make(env_id, render_mode="rgb_array")
        if seed is not None:
            env.seed(seed)
        self.simulator[name].update(
            env_id=env_id,
            env=env,
            buffer_obs=[],
            buffer_reward=None,
            buffer_terminated=None,
            buffer_truncated=None,
            next_action=None,
        )

    def pre_reset(self):
        pass

    @register.states()
    def reset(self):
        for _obj_name, sim in self.simulator.items():
            obs, info = sim["env"].reset()
            sim["buffer_obs"] = [obs]
            sim["buffer_reward"] = []
            sim["buffer_terminated"] = []
            sim["buffer_truncated"] = []

    def callback(self, t_n: float):
        for _obj_name, sim in self.simulator.items():
            next_action = sim["next_action"]
            obs, reward, is_terminated, truncated, _ = sim["env"].step(next_action)
            sim["buffer_obs"].append(obs)
            sim["buffer_reward"].append(reward)
            sim["buffer_terminated"].append(is_terminated)
            sim["buffer_truncated"].append(truncated)



from typing import Optional, List
import PIL
import numpy as np
import gymnasium as gym
from pyvirtualdisplay import Display
import os
import threading

# IMPORT EAGERX
import eagerx.core.register as register
from eagerx.core.space import Space
from eagerx.utils.utils import Msg
from eagerx.core.entities import EngineNode
from eagerx.core.constants import process


class ObservationSensor(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: Optional[int] = process.ENGINE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
    ):
        """ObservationSensor spec"""
        spec = cls.get_specification()

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if isinstance(inputs, list) else ["tick"]
        spec.config.outputs = outputs if isinstance(outputs, list) else ["observation"]
        return spec

    def initialize(self, spec, simulator):
        # We will probably use self.simulator in callback & reset.
        assert (
            self.process == process.ENGINE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Engine process"
        self.simulator = simulator
        self.id = simulator["env_id"]
        self.last_obs = None

    @register.states()
    def reset(self):
        self.last_obs = None

    @register.inputs(tick=Space(shape=(), dtype="int64"))
    @register.outputs(observation=Space(dtype="float32"))
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator, dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator
        )
        obs = self.simulator["buffer_obs"]
        self.simulator["buffer_obs"] = []
        obs = self.last_obs if len(obs) == 0 else obs[-1]
        self.last_obs = obs
        return dict(observation=obs)


class RewardSensor(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: Optional[int] = process.ENGINE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
    ):
        """RewardSensor spec"""
        spec = cls.get_specification()

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if isinstance(inputs, list) else ["tick"]
        spec.config.outputs = outputs if isinstance(outputs, list) else ["reward"]
        return spec

    def initialize(self, spec, simulator):
        # We will probably use self.simulator in callback & reset.
        assert (
            self.process == process.ENGINE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Engine process"
        self.simulator = simulator
        self.id = simulator["env_id"]
        self.last_reward = None

    @register.states()
    def reset(self):
        self.last_reward = 0.0

    @register.inputs(tick=Space(shape=(), dtype="int64"))
    @register.outputs(reward=Space(dtype="float32"))
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator, dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator
        )
        reward = self.simulator["buffer_reward"]
        self.simulator["buffer_reward"] = []
        if len(reward) == 0:
            reward = self.last_reward
        else:
            self.last_reward = reward[-1]
            reward = sum(reward)
        return dict(reward=reward)


class TruncatedSensor(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: Optional[int] = process.ENGINE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
    ):
        """DoneSensor spec"""
        spec = cls.get_specification()

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if isinstance(inputs, list) else ["tick"]
        spec.config.outputs = outputs if isinstance(outputs, list) else ["truncated"]
        return spec

    def initialize(self, spec, simulator):
        # We will probably use self.simulator in callback & reset.
        assert (
            self.process == process.ENGINE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Engine process"
        self.simulator = simulator
        self.id = simulator["env_id"]
        self.last_truncated = None

    @register.states()
    def reset(self):
        self.last_truncated = False

    @register.inputs(tick=Space(shape=(), dtype="int64"))
    @register.outputs(truncated=Space(low=0, high=1, shape=(), dtype="int64"))
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator, dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator
        )
        truncated = self.simulator["buffer_truncated"]
        self.simulator["buffer_truncated"] = []
        if len(truncated) == 0:
            truncated = self.last_truncated
        else:
            truncated = any(truncated) or self.last_truncated
            self.last_truncated = truncated
        return dict(truncated=int(truncated))


class TerminatedSensor(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: Optional[int] = process.ENGINE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
    ):
        """DoneSensor spec"""
        spec = cls.get_specification()

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if isinstance(inputs, list) else ["tick"]
        spec.config.outputs = outputs if isinstance(outputs, list) else ["terminated"]
        return spec

    def initialize(self, spec, simulator):
        # We will probably use self.simulator in callback & reset.
        assert (
            self.process == process.ENGINE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Engine process"
        self.simulator = simulator
        self.id = simulator["env_id"]
        self.last_terminated = None

    @register.states()
    def reset(self):
        self.last_terminated = False

    @register.inputs(tick=Space(shape=(), dtype="int64"))
    @register.outputs(terminated=Space(low=0, high=1, shape=(), dtype="int64"))
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator, dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator
        )
        terminated = self.simulator["buffer_terminated"]
        self.simulator["buffer_terminated"] = []
        if len(terminated) == 0:
            terminated = self.last_terminated
        else:
            terminated = any(terminated) or self.last_terminated
            self.last_terminated = terminated
        return dict(terminated=int(terminated))


class ActionActuator(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        zero_action=None,
        process: Optional[int] = process.ENGINE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "green",
        delay_state: bool = False,
    ):
        """ActionActuator spec"""
        spec = cls.get_specification()

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if isinstance(inputs, list) else ["tick", "action"]
        spec.config.outputs = outputs if isinstance(outputs, list) else ["action_applied"]
        spec.config.states = ["delay"] if delay_state else []

        # Modify custom node params
        spec.config.zero_action = zero_action
        return spec

    def initialize(self, spec, simulator):
        # We will probably use self.simulator in callback & reset.
        assert (
            self.process == process.ENGINE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Engine process"
        self.simulator = simulator
        self.simulator["env"]: gym.Env
        self.is_discrete = True if isinstance(self.simulator["env"].action_space, gym.spaces.Discrete) else False
        if spec.config.zero_action is None:
            self.zero_action = self.simulator["env"].action_space.sample()
        else:
            if isinstance(spec.config.zero_action, list):
                dtype = self.simulator["env"].action_space.dtype
                self.zero_action = np.array(spec.config.zero_action, dtype=dtype)
            else:
                self.zero_action = spec.config.zero_action
            assert self.simulator["env"].action_space.contains(
                self.zero_action
            ), f'The zero action provided for "{self.simulator["name"]}" is not contained in the action space of this environment.'

    @register.states(delay=Space(shape=(), low=0.0, high=0.0, dtype="float32"))
    def reset(self, delay=None):
        # This controller is stateless (in contrast to e.g. a PID controller).
        if delay is not None:
            self.set_delay(delay, "inputs", "action")
        self.simulator["next_action"] = self.zero_action

    @register.inputs(tick=Space(shape=(), dtype="int64"), action=Space(dtype="float32"))
    @register.outputs(action_applied=Space(dtype="float32"))
    def callback(
        self,
        t_n: float,
        tick: Optional[Msg] = None,
        action: Optional[Msg] = None,
    ):
        assert isinstance(self.simulator, dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator
        )

        # Set action in simulator for next step.
        if len(action.msgs) > 0:
            self.simulator["next_action"] = int(action.msgs[-1].item()) if self.is_discrete else action.msgs[-1]
        else:
            self.simulator["next_action"] = self.zero_action

        # Prepare output message
        action_applied = self.simulator["next_action"]

        # Send action that has been applied.
        return dict(action_applied=action_applied)


class GymImage(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: Optional[int] = process.ENGINE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
        shape: Optional[List[int]] = None,
        always_render=False,
    ):
        """GymImage spec"""
        spec = cls.get_specification()

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if isinstance(inputs, list) else ["tick"]
        spec.config.outputs = outputs if isinstance(outputs, list) else ["image"]

        # Modify custom node params
        spec.config.shape = shape if isinstance(shape, list) else [200, 200]
        spec.config.always_render = always_render
        spec.outputs.image.space = Space(low=0, high=255, shape=tuple(spec.config.shape + [3]), dtype="uint8")
        return spec

    def initialize(self, spec, simulator):
        # We will probably use self.simulator in callback & reset.
        assert (
            self.process == process.ENGINE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Engine process"
        self.simulator = simulator
        self.shape = tuple(spec.config.shape)
        self.always_render = spec.config.always_render
        self.render_toggle = False
        self.id = simulator["env_id"]
        self._cond = threading.Condition()
        self.sub_toggle = self.backend.Subscriber("%s/env/render/toggle" % self.ns, "bool", self._set_render_toggle)

        # Setup virtual display for rendering.
        self.display = Display(visible=False, backend="xvfb")
        self.disp_id = os.environ["DISPLAY"]  # First record default display id
        self.display.start()
        self.xvfb_id = self.display.env()["DISPLAY"]  # Get virtual display id
        os.environ["DISPLAY"] = self.disp_id  # Reset to default display id

    @register.states()
    def reset(self):
        # This sensor is stateless (in contrast to e.g. a Kalman filter).
        pass

    @register.inputs(tick=Space(shape=(), dtype="int64"))
    @register.outputs(image=Space(dtype="uint8"))
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator, dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator
        )
        with self._cond:
            if self.always_render or self.render_toggle:
                os.environ["DISPLAY"] = self.xvfb_id  # Set virtual display id
                try:
                    rgb = self.simulator["env"].render()
                except Exception as e:
                    self.backend.logwarn(e)
                    raise e
                os.environ["DISPLAY"] = self.disp_id  # Reset to default display id

                # Resize image if not matching desired self.shape
                if rgb.shape[:2] != tuple(self.shape):
                    img_PIL = PIL.Image.fromarray(rgb).convert("RGB")
                    img_PIL = img_PIL.resize(self.shape, PIL.Image.ANTIALIAS)
                    rgb = np.array(img_PIL).astype(rgb.dtype)
            else:
                rgb = np.zeros((self.shape[0], self.shape[1], 3), np.uint8)
        return dict(image=rgb)

    def shutdown(self):
        self.backend.logdebug(f"[{self.name}] {self.name}.shutdown() called.")
        self.sub_toggle.unregister()

    def _show_ros_image(self, msg):
        import matplotlib.pyplot as plt

        rgb_back = np.frombuffer(msg.data, dtype=np.uint8).reshape(msg.height, msg.width, -1)
        plt.imshow(rgb_back)
        plt.show()

    def _set_render_toggle(self, msg):
        with self._cond:
            if msg:
                self.backend.logdebug("[%s] START RENDERING!" % self.name)
            else:
                self.simulator["env"].close()
                self.backend.logdebug("[%s] STOPPED RENDERING!" % self.name)
            self.render_toggle = msg



from eagerx.core.space import Space
from eagerx.core.entities import Backend
from eagerx.core.specs import NodeSpec, EngineSpec, BackendSpec, BaseNodeSpec
from eagerx.core.entities import Node
from eagerx.core.graph import Graph
from eagerx.utils.node_utils import (
    initialize_nodes,
    wait_for_node_initialization,
)
from eagerx.core.executable_node import RxNode
from eagerx.core.executable_engine import RxEngine
from eagerx.core.supervisor import Supervisor, SupervisorNode
from eagerx.core.rx_message_broker import RxMessageBroker
from eagerx.core.constants import process
from eagerx.core.constants import ENVIRONMENT, NEW_PROCESS, EXTERNAL, BackendException

# OTHER IMPORTS
import copy
import atexit
import abc
import numpy as np
import functools
from typing import List, Union, Dict, Tuple, Optional, Any
import gymnasium as gym


class BaseEnv(gym.Env):
    """The base class for all EAGERx environments that follows the OpenAI gym's Env API.

    - Be sure to call :func:`super().__init__` inside the subclass' constructor with the required arguments (name, graph, etc...).

    A subclass should implement the following methods:

    - :func:`~eagerx.core.env.BaseEnv.step`: Be sure to call :func:`~eagerx.core.env.BaseEnv._step` inside this method to perform the step.

    - :func:`~eagerx.core.env.BaseEnv.reset`: Be sure to call :func:`~eagerx.core.env.BaseEnv._reset` inside this method to perform the reset.

    A subclass can optionally overwrite the following properties:

    - :attr:`~eagerx.core.env.BaseEnv.observation_space`: Per default, the observations, registered in the graph, are taken.

    - :attr:`~eagerx.core.env.BaseEnv.action_space`: Per default, the actions, registered in the graph, are taken.
    """

    def __init__(
        self,
        name: str,
        rate: float,
        graph: Graph,
        engine: EngineSpec,
        backend: BackendSpec = None,
        force_start: bool = True,
        render_mode: str = None,
    ) -> None:
        """Initializes  an environment with EAGERx dynamics.

        :param name: The name of the environment. Everything related to this environment
                     (parameters, topics, nodes, etc...) will be registered under namespace: "`/name`".
        :param rate: The rate (Hz) at which the environment will run.
        :param graph: The graph consisting of nodes and objects that describe the environment's dynamics.
        :param engine: The physics engine that will govern the environment's dynamics.
                       For every :class:`~eagerx.core.entities.Object` in the graph,
                       the corresponding engine implementations is chosen.
        :param backend: The backend that will govern the communication for this environment.
                        Per default, the :class:`~eagerx.backends.single_process.SingleProcess` backend is used.
        :param force_start: If there already exists an environment with the same name, the existing environment is
                            first shutdown by calling the :func:`~eagerx.core.env.BaseEnv` method before initializing this
                            environment.
        :param render_mode: The render mode that will be used for rendering the environment.
        """
        assert "/" not in name, 'Environment name "%s" cannot contain the reserved character "/".' % name
        self.name = name
        self.ns = "/" + name
        self.initialized = False
        self.has_shutdown = False
        self.render_mode = render_mode
        self._is_initialized = dict()
        self._launch_nodes = dict()
        self._sp_nodes = dict()

        # Register graph (returns unlinked specs with original graph & reloads entities).
        if engine is None:
            environment, engine, nodes, render = Graph._get_all_node_specs(graph._state)
            self.graph = graph
        else:
            self.graph, environment, engine, nodes, self.render_node = graph.register(rate, engine)
        self.rate = rate if rate else environment.config.rate

        # Initialize backend
        if backend is None:
            from eagerx.backends.single_process import SingleProcess

            backend = SingleProcess.make()
        self.backend = Backend.from_cmd(
            self.ns,
            backend.config.entity_id,
            backend.config.log_level,
            main=True,
            real_time_factor=engine.config.real_time_factor,
            sync=engine.config.sync,
            simulate_delays=engine.config.sync,
        )

        # Check if there already exists an environment
        self._shutdown_srv = self.backend.register_environment(self.ns, force_start, self.shutdown)

        # Delete pre-existing parameters
        self.backend.delete_param(f"/{self.name}", level=2)

        # Upload relevant run-time settings
        secs, nsecs = self.backend.serialize_time(self.backend.ts_init)
        self.backend.upload_params(
            self.ns,
            {
                "log_level": self.backend.log_level,
                "ts_init_secs": secs,
                "ts_init_nsecs": nsecs,
                "real_time_factor": self.backend.real_time_factor,
                "sync": self.backend.sync,
                "simulate_delays": self.backend.simulate_delays,
            },
        )

        # Initialize message broker
        self.mb = RxMessageBroker(owner="%s/%s" % (self.ns, "env"), backend=self.backend)

        # Create supervisor spec (adds addresses of (engine)states to supervisor)
        supervisor = self._create_supervisor(environment, engine, nodes)

        # Upload parameters
        self._upload_params(self.ns, self.backend, [supervisor, environment, engine] + nodes)

        # Initialize environment
        self.environment_node = RxNode(name="%s/%s" % (self.ns, environment.config.name), message_broker=self.mb)
        self.environment_node.node_initialized()
        self.environment = self.environment_node.node

        # Initialize nodes
        initialize_nodes(nodes, ENVIRONMENT, self.ns, self.mb, self._is_initialized, self._sp_nodes, self._launch_nodes)

        # Initialize engine
        initialize_nodes(
            engine,
            ENVIRONMENT,
            self.ns,
            self.mb,
            self._is_initialized,
            self._sp_nodes,
            self._launch_nodes,
            rxnode_cls=RxEngine,
        )

        # Initialize supervisor node
        self.supervisor_node = Supervisor("%s/%s" % (self.ns, supervisor.config.name), self.mb, self.environment)
        self.supervisor_node.node_initialized()
        self.supervisor = self.supervisor_node.node
        self.mb.connect_io()

        # Implement clean up
        atexit.register(self.shutdown)

    @staticmethod
    def _create_supervisor(environment: NodeSpec, engine: EngineSpec, nodes: List[NodeSpec]) -> BaseNodeSpec:
        entity_type = f"{SupervisorNode.__module__}/{SupervisorNode.__name__}"
        spec = Node.pre_make("N/a", entity_type)
        spec.add_output("step", space=Space(shape=(), dtype="int64"))

        spec.config.rate = environment.config.rate
        spec.config.name = "env/supervisor_node"
        spec.config.color = "yellow"
        spec.config.process = process.ENVIRONMENT
        spec.config.outputs = ["step"]

        # Get all states from all nodes
        for i in [environment, engine] + nodes:
            for cname in i.params["config"]["states"]:
                entity_name = i.config.name
                name = f"{entity_name}/{cname}"
                address = f"{entity_name}/states/{cname}"
                processor = None  # Only add processor (i.params["states"][cname]["processor"]) once at input side.
                space = i.params["states"][cname]["space"]

                assert (
                    name not in spec.params["states"]
                ), f'Cannot have duplicate states. State "{name}" is defined multiple times.'

                mapping = dict(address=address, processor=processor, space=space)
                with spec.states as d:
                    d[name] = mapping
                spec.config.states.append(name)

        # Get all engine states
        for entity_name, obj in engine.objects.items():
            for cname, state in obj.engine_states.items():
                name = f"{entity_name}/{cname}"
                address = f"{entity_name}/states/{cname}"
                processor = None  # Only add processor (i.params["states"][cname]["processor"]) once at input side.
                space = state["space"].to_dict()

                assert (
                    name not in spec.params["states"]
                ), f'Cannot have duplicate states. State "{name}" is defined multiple times.'

                mapping = dict(address=address, processor=processor, space=space)
                with spec.states as d:
                    d[name] = mapping
                spec.config.states.append(name)

        return spec

    @staticmethod
    def _upload_params(ns: str, backend: Backend, nodes: List[BaseNodeSpec]) -> None:
        for node in nodes:
            name = node.config.name
            msg = f"Node name '{ns}/{node.config.name}' already exists on the parameter server. Node names must be unique."
            assert backend.get_param(f"{ns}/rate/{name}", None) is None, msg
            # If no backend multiprocessing support, overwrite NEW_PROCESS to ENVIRONMENT
            if node.config.process == NEW_PROCESS and not backend.MULTIPROCESSING_SUPPORT:
                backend.logwarn_once(
                    f"Backend '{backend.BACKEND}' does not support multiprocessing, "
                    "so all nodes are launched in the ENVIRONMENT process."
                )
                node.config.process = ENVIRONMENT
            # Raise error if external process is not supported
            elif node.config.process == EXTERNAL and not backend.DISTRIBUTED_SUPPORT:
                raise BackendException(
                    f"Backend '{backend.BACKEND}' does not support distributed computation. "
                    f"Therefore, this backend is incompatible with node '{name}', "
                    f"because {name}.config.process=EXTERNAL."
                )
            params = node.build(ns=ns)
            backend.upload_params(ns, params)

    @property
    def observation_space(self) -> gym.spaces.Space:
        """The Space object corresponding to valid observations.

        Per default, the observation space of all registered observations in the graph is used.
        """
        return self._observation_space

    @property
    def action_space(self) -> gym.spaces.Space:
        """The Space object corresponding to valid actions

        Per default, the action space of all registered actions in the graph is used.
        """
        return self._action_space

    @property
    def state_space(self) -> gym.spaces.Dict:
        """Infers the state space from the space of every state.

        This space defines the format of valid states that can be set before the start of an episode.

        :returns: A dictionary with *key* = *state* and *value* = :class:`Space`.
        """
        state_space = dict()
        for name, buffer in self.supervisor.state_buffer.items():
            state_space[name] = buffer["space"]
        return gym.spaces.Dict(spaces=state_space)

    @property
    def _observation_space(self) -> gym.spaces.Dict:
        """Infers the observation space from the space of every observation.

        This space defines the format of valid observations.

        .. note:: Observations with :attr:`~eagerx.core.specs.RxInput.window` = 0 are excluded from the observation space.
                  For observations with :attr:`~eagerx.core.specs.RxInput.window` > 1,
                  the observation space is duplicated :attr:`~window` times.

        :returns: A dictionary with *key* = *observation* and *value* = :class:`Space`.
        """
        assert not self.has_shutdown, "This environment has been shutdown."
        observation_space = dict()
        for name, buffer in self.environment.observation_buffer.items():
            space = buffer["space"]
            if not buffer["window"] > 0:
                continue
            if isinstance(space, gym.spaces.Discrete):
                stacked_space = gym.spaces.MultiDiscrete([space.n] * buffer["window"])
            else:
                low = np.repeat(space.low[np.newaxis, ...], buffer["window"], axis=0)
                high = np.repeat(space.high[np.newaxis, ...], buffer["window"], axis=0)
                stacked_space = gym.spaces.Box(low=low, high=high, dtype=space.dtype)
            observation_space[name] = stacked_space
        return gym.spaces.Dict(spaces=observation_space)

    @property
    def _action_space(self) -> gym.spaces.Dict:
        """Infers the action space from the space of every action.

        This space defines the format of valid actions.

        :returns: A dictionary with *key* = *action* and *value* = :class:`Space`.
        """
        assert not self.has_shutdown, "This environment has been shutdown."
        action_space = dict()
        for name, buffer in self.environment.action_buffer.items():
            action_space[name] = buffer["space"]
        return gym.spaces.Dict(spaces=action_space)

    def _set_action(self, action) -> None:
        # Set actions in buffer
        for name, buffer in self.environment.action_buffer.items():
            assert not self.supervisor.sync or name in action, (
                'Action "%s" not specified. Must specify all actions in action_space if running reactive.' % name
            )
            if name in action:
                buffer["msg"] = action[name]

    def _set_state(self, state) -> None:
        # Set states in buffer
        for name, msg in state.items():
            assert name in self.supervisor.state_buffer, 'Cannot set unknown state "%s".' % name
            self.supervisor.state_buffer[name]["msg"] = msg

    def _get_observation(self) -> Dict:
        # Get observations from buffer
        observation = dict()
        for name, buffer in self.environment.observation_buffer.items():
            observation[name] = buffer["msgs"]
        return observation

    def _initialize(self, states: Dict) -> None:
        assert not self.initialized, "Environment already initialized. Cannot re-initialize pipelines. "

        # Set desired reset states
        self._set_state(states)

        # Wait for nodes to be initialized
        [node.node_initialized() for name, node in self._sp_nodes.items()]
        wait_for_node_initialization(self._is_initialized, self.backend)

        # Initialize communication within this process
        self.mb.connect_io(print_status=True)

        self.backend.logdebug("Nodes initialized.")

        # Perform first reset
        self.supervisor.reset()

        # Nodes initialized
        self.initialized = True
        self.backend.loginfo("Communication initialized.")

    def _shutdown(self):
        if not self.has_shutdown:
            self._shutdown_srv.unregister()
            for address, node in self._launch_nodes.items():
                self.backend.logdebug(f"[{self.name}] Send termination signal to '{address}'.")
                node.terminate()
            for _, rxnode in self._sp_nodes.items():
                rxnode: RxNode
                if not rxnode.has_shutdown:
                    self.backend.logdebug(f"[{self.name}][{rxnode.name}] Shutting down.")
                    rxnode.node_shutdown()
            if not self.supervisor_node.has_shutdown:
                self.supervisor_node.node_shutdown()
            if not self.environment_node.has_shutdown:
                self.environment_node.node_shutdown()
            self.mb.shutdown()
            self.backend.delete_param(f"/{self.name}", level=1)
            self.backend.shutdown()
            self.has_shutdown = True

    def _reset(self, states: Dict) -> Dict:
        """A private method that should be called within :func:`~eagerx.core.env.BaseEnv.reset()`.

        :param states: The desired states to be set before the start an episode.
                       May also be an (empty) subset of registered states if not all states require a reset.
        :returns: The initial observation.
        """
        assert not self.has_shutdown, "This environment has been shutdown."
        # Initialize environment
        if not self.initialized:
            self._initialize(states)

        # Set desired reset states
        self._set_state(states)

        # Perform reset
        self.supervisor.reset()
        obs = self._get_observation()
        return obs

    @abc.abstractmethod
    def reset(self, seed: int = None, options: Dict[str, Any] = None) -> Tuple[Union[Dict, np.ndarray], Dict]:
        """An abstract method that resets the environment to an initial state and returns an initial observation.

        .. note:: To reset the graph, the private method :func:`~eagerx.core.env.BaseEnv._reset` must be called with the
                   desired initial states. The spaces of all states (of Objects and Nodes in the graph) are stored in
                  :func:`~eagerx.core.env.BaseEnv.state_space`.

        :returns: The initial observation that is complies with the :func:`~eagerx.core.env.BaseEnv.observation_space`.
        """
        pass

    def _step(self, action: Dict) -> Dict:
        """A private method that should be called within :func:`~eagerx.core.env.BaseEnv.step()`.

        :param action: The actions to be applied in the next timestep.
                       Should include all registered actions.
        :returns: The observation of the current timestep that comply with the graph's observation space.
        """
        # Check that nodes were previously initialized.
        assert self.initialized, "Not yet initialized. Call .reset() before calling .step()."
        assert not self.has_shutdown, "This environment has been shutdown."

        # Set actions in buffer
        self._set_action(action)

        # Call step
        self.supervisor.step()
        return self._get_observation()

    @abc.abstractmethod
    def step(self, action: Union[Dict, np.ndarray]) -> Tuple[Union[Dict, np.ndarray], float, bool, bool, Dict]:
        """An abstract method that runs one timestep of the environment's dynamics.

        .. note:: To run one timestep of the graph dynamics (that essentially define the environment dynamics),
                  this method must call the private method :func:`~eagerx.core.BaseEnv._step` with the actions that comply
                  with :attr:`~eagerx.core.BaseEnv._action_space`.

        When the end of an episode is reached, the user is responsible for calling :func:`~eagerx.core.BaseEnv.reset`
        to reset this environment's state.

        :params action: Actions provided by the agent. Should comply with the :func:`~eagerx.core.env.BaseEnv.action_space`.

        :returns: A tuple (observation, reward, terminated, truncated, info).

                  - observation: Observations of the current timestep that comply with
                                 the :func:`~eagerx.core.env.BaseEnv.observation_space`.

                  - reward: amount of reward returned after previous action

                  - terminated: whether the episode has ended due to a terminal state, in which case further step() calls will
                                return undefined results

                  - truncated: whether the episode has ended due to a time limit, in which case further step() calls will
                               return undefined results

                  - info: contains auxiliary diagnostic information (helpful for debugging, and sometimes learning)
        """
        pass

    @functools.wraps(Graph.gui)
    def gui(
        self,
        *args,
        interactive: Optional[bool] = True,
        resolution: Optional[List[int]] = None,
        filename: Optional[str] = None,
        **kwargs,
    ) -> Union[None, np.ndarray]:
        """Opens a graphical user interface of the graph that was used to initialize this environment.

        .. note:: Requires `eagerx-gui`:

                .. highlight:: python
                .. code-block:: python

                    pip3 install eagerx-gui

        :param interactive: If `True`, an interactive application is launched.
                            Otherwise, an RGB render of the GUI is returned.
                            This could be useful when using a headless machine.
        :param resolution: Specifies the resolution of the returned render when `interactive` is `False`.
                           If `interactive` is `True`, this argument is ignored.
        :param filename: If provided, the GUI is rendered to an svg file with this name.
                         If `interactive` is `True`, this argument is ignored.
        :return: RGB render of the GUI if `interactive` is `False`.
        """
        return self.graph.gui(*args, interactive=interactive, resolution=resolution, filename=filename, **kwargs)

    def save(self, file: str) -> None:
        """Saves the (engine-specific) graph state, that includes the engine & environment nodes.

        The state is saved in *.yaml* format and contains the state of every added node, action, and observation
        and the connections between them.

        :param file: A string giving the name (and the file if the file isn't in the current working directory).
        """
        return self.graph.save(file)

    @classmethod
    def load(cls, name: str, file: str, backend: BackendSpec = None, force_start: bool = True):
        """Loads an environment corresponding to the graph state.

        :param name: The name of the environment. Everything related to this environment
             (parameters, topics, nodes, etc...) will be registered under namespace: "`/name`".
        :param file: A string giving the name (and the file if the file isn't in the current working directory).
        :param backend: The backend that will govern the communication for this environment.
                        Per default, the :class:`~eagerx.backends.single_process.SingleProcess` backend is used.
        :param force_start: If there already exists an environment with the same name, the existing environment is
                            first shutdown by calling the :func:`~eagerx.core.env.BaseEnv` method before initializing this
                            environment.
        """
        graph = Graph.load(file)
        graph = copy.deepcopy(graph)
        return cls(name=name, rate=None, graph=graph, engine=None, backend=backend, force_start=force_start)

    def render(self) -> Optional[np.ndarray]:
        """A method to start rendering (i.e. open the render window).

        A bool message to topic address ":attr:`~eagerx.core.env.BaseEnv.name` */env/render/toggle*",
        which toggles the rendering on/off.
        :returns: Optionally, a rgb_array if env.mode=rgb_array.
        """
        mode = self.render_mode
        assert not self.has_shutdown, "This environment has been shutdown."
        if self.render_node:
            if mode == "human":
                self.supervisor.start_render()
            elif mode == "rgb_array":
                self.supervisor.start_render()
                img = self.supervisor.get_last_image()
                return img
            elif mode is None:
                return
            else:
                raise ValueError('Render mode "%s" not recognized.' % mode)
        else:
            self.backend.logwarn_once("No render node active, so not rendering.")
            if mode == "rgb_array":
                return np.empty((0, 0, 3), dtype="uint8")
            else:
                return

    def close(self):
        """A method to stop rendering (i.e. close the render window).

        A bool message to topic address ":attr:`~eagerx.core.env.BaseEnv.name` */env/render/toggle*",
        which toggles the rendering on/off.

        .. note:: Depending on the source node that is producing the images that are rendered,
                  images may still be produced, even when the render window is not visible.
                  This may add computational overhead and influence the run speed.

                  Optionally, users may subscribe to topic address ":attr:`~eagerx.core.env.BaseEnv.name` */env/render/toggle*"
                  in the node that is producing the images to stop the production and output empty images instead.
        """
        assert not self.has_shutdown, "This environment has been shutdown."
        self.supervisor.stop_render()

    def shutdown(self):
        """A method to shutdown the environment.

        - Clear the parameters on the ROS parameter under the namespace /:attr:`~eagerx.core.env.BaseEnv.name`.

        - Close nodes (i.e. release resources and perform :class:`~eagerx.core.entities.Node.close` procedure).

        - Unregister topics that supplied the I/O communication between nodes.
        """
        self._shutdown()



from typing import List, Dict, Optional, Union, Any, Callable, Tuple
import abc
import inspect
import os
import time
import numpy as np
import psutil
from tabulate import tabulate
from termcolor import colored

from eagerx.core.log import frame_to_caller_id, LoggingOnce
import eagerx
from eagerx.core.pubsub import Publisher, Subscriber, ShutdownService
from eagerx.core.specs import ObjectSpec
from eagerx.core.constants import (
    ENGINE,
    SILENT,
    DEBUG,
    INFO,
    ERROR,
    WARN,
    FATAL,
    Unspecified,
)
from eagerx.core.rx_message_broker import RxMessageBroker
from eagerx.utils.node_utils import initialize_nodes, wait_for_node_initialization
from eagerx.utils.utils import Msg, load, initialize_processor, get_param_with_blocking

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from eagerx.core.graph_engine import EngineGraph  # noqa: F401
    from eagerx.core.specs import (  # noqa:
        EntitySpec,
        EngineSpec,
        NodeSpec,
        ProcessorSpec,
        EngineStateSpec,
        ResetNodeSpec,
        ObjectSpec,
        BackendSpec,
    )


# A singleton that is used to check if an argument was specified.
_unspecified = Unspecified()


class Entity(object):
    @classmethod
    def get_specification(cls):
        entity_type = cls.__module__ + "/" + cls.__qualname__
        spec = cls.pre_make(entity_type, entity_type)
        spec.initialize(cls)
        return spec

    @classmethod
    @abc.abstractmethod
    def make(cls, *args: Any, **kwargs: Any):
        """An abstract method that makes the specification (also referred to as `spec`) of this entity.

        :param args: Arguments to the subclass' make function.
        :param kwargs: Optional Arguments to the subclass' make function.
        :return: A (mutable) spec that can be used to build and subsequently initialize the entity (e.g. node, engine, ...).
        """
        pass

    @classmethod
    def info(cls, method: Optional[Union[List[str], str]] = None) -> str:
        """A helper method to get info on a method of the specified subclass.

        :param method: The registered method we would like to receive info on. If no method is specified, it provides info on
                       the class itself.
        :return: Info on the subclass' method.
        """
        from eagerx.core import info

        return info.get_info(cls, method)

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        from eagerx.core.specs import EntitySpec  # noqa: F811 F401

        return EntitySpec(dict(entity_id=entity_id, entity_type=entity_type))

    @classmethod
    def check_spec(cls, spec):
        pass


class Backend(Entity):
    """Baseclass for backends.

    Use this baseclass to implement backends that implement the communication.

    Users must use :func:`~eagerx.core.entities.Backend.make` to make the registered subclass' specification.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.Backend.make`

    - :func:`~eagerx.core.entities.Backend.initialize`

    - :func:`~eagerx.core.entities.Backend.Publisher`

    - :func:`~eagerx.core.entities.Backend.Subscriber`

    - :func:`~eagerx.core.entities.Backend.register_environment`

    - :func:`~eagerx.core.entities.Backend.delete_param`

    - :func:`~eagerx.core.entities.Backend.upload_params`

    - :func:`~eagerx.core.entities.Backend.get_param`

    - :func:`~eagerx.core.entities.Backend.spin`

    Subclasses must set the following static class properties:

    - :attr:`~eagerx.core.entities.Backend.BACKEND`

    - :attr:`~eagerx.core.entities.Backend.DISTRIBUTED_SUPPORT`

    - :attr:`~eagerx.core.entities.Backend.MULTIPROCESSING_SUPPORT`

    - :attr:`~eagerx.core.entities.Backend.COLAB_SUPPORT`
    """

    INFO = {
        "make": [],
        "initialize": [],
    }

    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        ns: str,
        backend_type: str,
        entity_id: str,
        log_level: int,
        main: bool = False,
        sync: Optional[bool] = None,
        real_time_factor: Optional[float] = None,
        simulate_delays: Optional[bool] = None,
        **kwargs: Union[bool, int, float, str, List, Dict],
    ):
        #: Namespace of the environment. Can be set with the `name` argument to :class:`~eagerx.core.env.BaseEnv`.
        self.ns: str = ns
        #: A unique entity_id with the structure <module>/<classname>.
        self.entity_id: str = entity_id
        #: The class definition of the subclass. Follows naming convention *<module>/<BackendClassName>*.
        #: Cannot be modified.
        self.backend_type: str = backend_type
        #: If True, the backend is the 'main` backend that corresponds to the environment process.
        self.main: bool = main
        #: Specifies the effective log level:
        #: {0: SILENT, 10: DEBUG, 20: INFO, 30: WARN, 40: ERROR, 50: FATAL}.
        #: Can be set in the subclass' :func:`~eagerx.core.entities.Backend.spec`.
        self.log_level: int = log_level

        # Record once logged messages.
        self._logging_once = LoggingOnce()

        # Backend shutdown flag
        self._has_shutdown = False

        # Call subclass' initialize method
        self.initialize(**kwargs)

        if main:
            self.ts_init = time.time()
            t_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.ts_init / 1e9))
            self.loginfo(f"Environment '{self.ns}' initialized at {t_str}")
            self.sync = sync
            self.real_time_factor = real_time_factor
            self.simulate_delays = simulate_delays
        else:
            # Happens *after* initialization, else param server not yet online.
            #: Timestamp of when the environment was initialized.
            secs = get_param_with_blocking(f"{ns}/ts_init_secs", self, None)
            nsecs = get_param_with_blocking(f"{ns}/ts_init_nsecs", self, None)
            self.ts_init: float = self.deserialize_time(secs, nsecs)
            #: Flag that specifies whether we run synchronous or asynchronous.
            self.sync: bool = get_param_with_blocking(f"{ns}/sync", self, None)
            #: A specified upper bound on the real_time factor. Wall-clock rate=real_time_factor*rate.
            #: If real_time_factor < 1 the simulation is slower than real time.
            self.real_time_factor: float = get_param_with_blocking(f"{ns}/real_time_factor", self, None)
            #: Flag that specifies whether input delays are simulated.
            #: You probably want to set this to False when running in the real-world.
            self.simulate_delays: bool = get_param_with_blocking(f"{ns}/real_time_factor", self, None)

    @abc.abstractmethod
    def initialize(self, spec: "BackendSpec") -> None:
        """An abstract method to initialize the backend.

        :param spec: Specification of the node/engine.
        """
        pass

    @classmethod
    def pre_make(cls, entity_id: str, entity_type: "Entity"):
        spec = super().pre_make(entity_id, entity_type)
        params = spec.params
        params["config"] = dict(
            log_level=eagerx.get_log_level(),
            entity_id=params.pop("entity_id"),
        )
        params["backend_type"] = params.pop("entity_type")
        from eagerx.core.specs import BackendSpec  # noqa: F811

        return BackendSpec(params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)

    @staticmethod
    def from_cmd(ns: str, backend_id: str, log_level: int, *args, **kwargs) -> "Backend":
        backend_cls = load(backend_id)
        backend = backend_cls(ns, backend_id, backend_id, log_level, *args, **kwargs)
        return backend

    @property
    @abc.abstractmethod
    def BACKEND(self) -> str:
        """Backend name in string format."""
        pass

    @property
    @abc.abstractmethod
    def DISTRIBUTED_SUPPORT(self) -> bool:
        """Whether nodes can be launched on external platforms (i.e. distributed communication)."""
        pass

    @property
    @abc.abstractmethod
    def MULTIPROCESSING_SUPPORT(self) -> bool:
        """Whether nodes can be launched as separate processes."""
        pass

    @property
    @abc.abstractmethod
    def COLAB_SUPPORT(self) -> bool:
        """Whether the backend supports running on Google colab."""
        pass

    @abc.abstractmethod
    def Publisher(self, address: str, dtype: str) -> Publisher:
        """Creates a publisher.

        :param address: Topic name.
        :param dtype: Dtype of message in string format (e.g. `float32`).
        """
        pass

    @abc.abstractmethod
    def Subscriber(
        self, address: str, dtype: str, callback, header: bool = False, callback_args: Optional[Tuple] = tuple()
    ) -> Subscriber:
        """Creates a subscriber.

        :param address: Topic name.
        :param dtype: Dtype of message in string format (e.g. `float32`).
        :param callback: Function to call ( fn(data)) when data is received. If callback_args is set, the function
                         must accept the callback_args as positional args, i.e. fn(data, header, *callback_args).
        :param header: Set to True if the callback accepts the header as the second positional argument.
        :param callback_args: Additional arguments to pass to the callback.
        """
        pass

    @abc.abstractmethod
    def register_environment(self, name: str, force_start: bool, fn: Callable) -> ShutdownService:
        """Checks if environment already exists and shuts it down if `force_restart` is set. Then, it registers
        the remote shutdown procedure for the newly created environment.

        :param name: Environment name (i.e. namespace of the environment).
        :param force_start: Whether to shutdown any environment with the same name if it already exists.
        :param fn: Function with zero args to be called on remote shutdown.
        """
        pass

    @abc.abstractmethod
    def delete_param(self, param: str, level: int = 1) -> None:
        """Deletes params from the parameter server.

        :param param: Parameter name.
        :param level: Determines what to do when the param does not exist:

                      - 0=error: Raises a BackendException.

                      - 1=warn: logs a warning and returns `None`.

                      - 2=pass: passes silently and returns `None`.
        """
        pass

    @abc.abstractmethod
    def upload_params(
        self, ns: str, values: Dict[str, Union[Dict, List, bool, float, int, str]], verbose: bool = False
    ) -> None:
        """Upload params to the parameter server.

        :param ns: Namespace to load parameters into, ``str``.
        :param values: Key/value dictionary, where keys are parameter names and values are parameter values, ``dict``.
        :param verbose: Verbosity level.
        """
        pass

    @abc.abstractmethod
    def get_param(self, name: str, default: Any = _unspecified) -> Union[Dict, List, bool, float, int, str]:
        """Retrieve a parameter from the param server

        :param name: Parameter name.
        :param default: Default value to return.
        """
        pass

    @abc.abstractmethod
    def spin(self) -> None:
        """Blocks until node is shutdown. Yields activity to other threads."""
        pass

    def now(self) -> Tuple[float, float]:
        """Get the current times according to the simulated and wall clock"""
        wc = time.time()
        _passed = wc - self.ts_init
        sc = wc if self.real_time_factor == 0 else self.ts_init + _passed * self.real_time_factor
        return sc, wc

    @staticmethod
    def serialize_time(t: float) -> Tuple[int, int]:
        """
        Convert a float time instance (in seconds) into secs and nsecs.

        Should be used when manually setting secs/nsecs slot values for serialization.
        """
        secs = int(t)
        nsecs = int((t - secs) * 1e9)
        return secs, nsecs

    @staticmethod
    def deserialize_time(secs: int, nsecs: int) -> float:
        """
        Convert a secs and nsecs time instance into float time in seconds .

        Should be used when manually setting secs/nsecs slot values for deserialization.
        """
        return float(secs) + float(nsecs) / 1e9

    def shutdown(self) -> None:
        """Shuts down the backend"""
        if not self._has_shutdown:
            self._has_shutdown = True
            self.logdebug("Backend.shutdown() called.")

    def logdebug_once(self, msg, identifier=None) -> None:
        caller_id = frame_to_caller_id(inspect.currentframe().f_back.f_back)
        if isinstance(identifier, str):
            caller_id += str.encode(identifier)
        if self._logging_once(caller_id):
            return self.logdebug(msg)

    def loginfo_once(self, msg, identifier=None) -> None:
        caller_id = frame_to_caller_id(inspect.currentframe().f_back.f_back)
        if isinstance(identifier, str):
            caller_id += str.encode(identifier)
        if self._logging_once(caller_id):
            return self.loginfo(msg)

    def logwarn_once(self, msg, identifier=None) -> None:
        caller_id = frame_to_caller_id(inspect.currentframe().f_back)
        if isinstance(identifier, str):
            caller_id += str.encode(identifier)
        if self._logging_once(caller_id):
            return self.logwarn(msg)

    def logerr_once(self, msg, identifier=None) -> None:
        caller_id = frame_to_caller_id(inspect.currentframe().f_back.f_back)
        if isinstance(identifier, str):
            caller_id += str.encode(identifier)
        if self._logging_once(caller_id):
            return self.logerr(msg)

    def logfatal_once(self, msg, identifier=None) -> None:
        caller_id = frame_to_caller_id(inspect.currentframe().f_back.f_back)
        if isinstance(identifier, str):
            caller_id += str.encode(identifier)
        if self._logging_once(caller_id):
            return self.logfatal(msg)

    def logdebug(self, msg) -> None:
        return self._log(f"[DEBUG]: {msg}", DEBUG, None)

    def loginfo(self, msg) -> None:
        return self._log(f"[INFO]: {msg}", INFO, None)

    def logwarn(self, msg) -> None:
        return self._log(f"[WARN]: {msg}", WARN, "red")

    def logerr(self, msg) -> None:
        return self._log(f"[ERROR]: {msg}", ERROR, "red")

    def logfatal(self, msg: str) -> None:
        return self._log(f"[FATAL]: {msg}", FATAL, "red")

    def _log(self, msg: str, level: int, color: Union[str, None]) -> None:
        if level >= self.log_level:
            print(colored(msg, color))

    def get_log_fn(self, log_level: int) -> Callable:
        if log_level == DEBUG:
            return self.logdebug
        elif log_level == INFO:
            return self.loginfo
        elif log_level == WARN:
            return self.logwarn
        elif log_level == ERROR:
            return self.logerr
        elif log_level == FATAL:
            return self.logfatal
        else:

            def logsilent(msg, *args, **kwargs):
                pass

            return logsilent


class BaseNode(Entity):
    def __init__(
        self,
        ns: str,
        message_broker: RxMessageBroker,
        sync: bool,
        real_time_factor: float,
        simulate_delays: bool,
        params: Dict,
    ):
        """
        The base class from which all nodes and engines inherit.
        """
        #: Namespace of the environment. Can be set with the `name` argument to :class:`~eagerx.core.env.BaseEnv`.
        self.ns: str = ns
        #: User specified node name. Can be set in :func:`~eagerx.core.entities.Node.spec`.
        self.name: str = params["config"]["name"]
        #: User specified node name with the namespace prepended. Cannot be modified directly.
        self.ns_name: str = f"{ns}/{self.name}"
        #: Responsible for all I/O communication within this process.
        #: Nodes inside the same process share the same backend. Cannot be modified.
        self.message_broker: RxMessageBroker = message_broker
        #: Responsible for all I/O communication within this process.
        #: Nodes inside the same process share the same message broker. Cannot be modified.
        self.backend: Backend = message_broker.backend
        #: A unique entity_id with the structure <module>/<classname>.
        self.entity_id: str = params["config"]["entity_id"]
        #: Rate (Hz) at which the callback is called.
        #: Can be set in the subclass' :func:`~eagerx.core.entities.Node.spec`.
        self.rate: float = params["config"]["rate"]
        #: Process in which this node is launched. See :class:`~eagerx.core.constants.process` for all options.
        #: Can be set in the subclass' :func:`~eagerx.core.entities.Node.spec`.
        self.process: int = params["config"]["process"]
        #: Executable script that is used to launch the node (if launched in a separate process).
        #: Can be set in the subclass' :func:`~eagerx.core.entities.Node.spec`.
        self.executable: str = params["config"]["executable"]
        #: Parameters for all selected inputs.
        self.inputs: dict = params["inputs"]
        #: Parameters for all selected outputs.
        self.outputs: dict = params["outputs"]
        #: Parameters for all selected states.
        self.states: dict = params["states"]
        #: Parameters for all feedthroughs corresponding to the selected outputs.
        self.feedthroughs: dict = params["feedthroughs"]
        #: Parameters for all selected targets.
        self.targets: dict = params["targets"]
        #: Flag that specifies whether we run reactive or asynchronous.
        #: Can be set in the engine's :func:`~eagerx.core.entities.Engine.spec`.
        self.sync: bool = sync
        #: A specified upper bound on the real_time factor. Wall-clock rate=real_time_factor*rate.
        #: If real_time_factor < 1 the simulation is slower than real time.
        #: Can be set in the engine's :func:`~eagerx.core.entities.Engine.spec`.
        self.real_time_factor: float = real_time_factor
        #: Flag that specifies whether input delays are simulated.
        #: You probably want to set this to False when running in the real-world.
        #: Can be set in the engine's :func:`~eagerx.core.entities.Engine.spec`.
        self.simulate_delays: bool = simulate_delays
        #: Specifies the color of logged messages & node color in the GUI.
        #: Check-out the termcolor documentation for the supported colors.
        #: Can be set in the subclass' :func:`~eagerx.core.entities.Node.spec`.
        self.color: str = params["config"]["color"]
        #: Specifies the log level for this node:
        #: {0: SILENT, 10: DEBUG, 20: INFO, 30: WARN, 40: ERROR, 50: FATAL}.
        #: Can be set in the subclass' :func:`~eagerx.core.entities.Node.spec`.
        self.log_level: int = params["config"]["log_level"]
        effective_log_level = eagerx.get_log_level()
        #: Specifies the log level for logging memory usage over time for this node:
        #: {0: SILENT, 10: DEBUG, 20: INFO, 30: WARN, 40: ERROR, 50: FATAL}.
        #: Note that `log_level` has precedent over the memory level set here.
        #: Can be set in the subclass' :func:`~eagerx.core.entities.Node.spec`.
        self.log_memory: int = self.log_level >= effective_log_level and SILENT >= effective_log_level

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        spec = super().pre_make(entity_id, entity_type)
        params = spec.params
        params["node_type"] = params.pop("entity_type")
        params["config"] = dict(
            name=None,
            rate=None,
            process=0,
            inputs=[],
            outputs=[],
            states=[],
            color="grey",
            log_level=WARN,
            log_level_memory=SILENT,
            executable=None,
            entity_id=params.pop("entity_id"),
        )
        params.update(dict(inputs=dict(), outputs=dict(), states=dict()))
        from eagerx.core.specs import BaseNodeSpec

        return BaseNodeSpec(params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)
        entity_id = spec.config.entity_id
        name = spec.config.name
        assert name is not None and isinstance(
            name, str
        ), f'A node with entity_id "{entity_id}" has an invalid name {name}. Please provide a unique name of type string.'

        # Check that all selected cnames have a corresponding implementation
        for component in ["inputs", "outputs", "states"]:
            for cname in spec.config[component]:
                c = getattr(spec, component)
                msg = (
                    f'"{cname}" was selected as {component[:-1]} '
                    f'but it is not a registered {component[:-1]} of node "{name}". Check the spec of "{entity_id}".'
                )
                assert cname in c, msg

        # Check that components have a space
        for component in ["states"]:
            for cname, params in getattr(spec, component).items():
                msg = (
                    f'"{cname}" was defined as {component[:-1]} in "{name}", but its space ({type(params.space)}) '
                    f'is not a valid space. Check the spec of "{entity_id}".'
                )
                assert params.space is not None, msg

    @abc.abstractmethod
    def initialize(self, spec: Union["NodeSpec", "EngineSpec", "ResetNodeSpec"]) -> None:
        """An abstract method that initializes the node at run-time.

        :param spec: Specification of the node/engine.
        """
        pass

    def shutdown(self) -> None:
        """A method that can be overwritten to cleanly shutdown (e.g. release resources)."""
        pass


class Node(BaseNode):
    """Baseclass for nodes.

    Use this baseclass to implement nodes that will be added to the (agnostic) :class:`~eagerx.core.graph.Graph`.

    Users must call :func:`~eagerx.core.entities.Node.make` to make the node subclass' specification.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.Node.make`

    - :func:`~eagerx.core.entities.Node.initialize`

    - :func:`~eagerx.core.entities.Node.reset`

    - :func:`~eagerx.core.entities.Node.callback`

    - :func:`~eagerx.core.entities.Node.shutdown` (optional)

    Use baseclass :class:`~eagerx.core.entities.EngineNode` instead, for nodes that will be added to
    :class:`~eagerx.core.graph_engine.EngineGraph` when specifying an engine implementation for an :class:`~eagerx.core.entities.Object`.

    Use baseclass :class:`~eagerx.core.entities.ResetNode` instead, for reset routines.
    """

    INFO = {
        "make": [],
        "initialize": [],
        "reset": ["states"],
        "callback": ["inputs", "outputs"],
    }

    def __init__(
        self,
        ns: str,
        message_broker: RxMessageBroker,
        sync: bool,
        real_time_factor: float,
        simulate_delays: bool,
        params: Dict,
        call_init: bool = True,
    ):

        # Message counter
        self.num_ticks = 0

        # Track memory usage  & speed
        self.total_ticks = 0
        self.py = psutil.Process(os.getpid())
        self.pid = os.getpid()
        self.iter_start = None
        self.iter_ticks = 0
        self.print_iter = 200
        self.history = []
        self.headers = [
            "pid",
            "node",
            "ticks",
            "rss",
            "diff",
            "t0",
            "vms",
            "diff",
            "t0",
            "iter_time",
            "diff",
            "t0",
        ]

        # Call baseclass constructor (which eventually calls .initialize())
        super().__init__(
            ns=ns,
            message_broker=message_broker,
            sync=sync,
            real_time_factor=real_time_factor,
            simulate_delays=simulate_delays,
            params=params,
        )

        if call_init:
            from eagerx.core.specs import NodeSpec

            self.initialize(NodeSpec(params))

        # Determine all inputs with window > 0 (and skip)
        self.skipped_cbs = 0
        self.windowed = dict()
        for cname, i in self.inputs.items():
            window = i["window"]
            skip = i["skip"]
            if window > 0:
                self.windowed[cname] = skip

        # If we run async *and* no msg for input with window > 0, then send None outputs.
        self.empty_outputs = dict()
        for cname, _ in self.outputs.items():
            self.empty_outputs[cname] = None
        for cname, _ in self.targets.items():
            self.empty_outputs[cname + "/done"] = False

    def reset_cb(self, **kwargs):
        self.num_ticks = 0
        self.skipped_cbs = 0
        keys_to_pop = []
        for cname, msg in kwargs.items():
            if msg.info.done:
                keys_to_pop.append(cname)
            else:
                kwargs[cname] = msg.msgs[0]
        [kwargs.pop(key) for key in keys_to_pop]
        return self.reset(**kwargs)

    def callback_cb(self, node_tick: int, t_n: float, **kwargs):
        self.iter_ticks += 1
        if self.log_memory and self.iter_ticks % self.print_iter == 0:
            if self.iter_start:
                # Time steps
                iter_stop = time.time()
                if self.iter_ticks > 0:
                    iter_time = float((iter_stop - self.iter_start) / self.iter_ticks)
                else:
                    iter_time = float(iter_stop - self.iter_start)

                # Memory usage request
                mem_use = (np.array(self.py.memory_info()[0:2]) / 2.0**30) * 1000  # memory use in MB...I think

                # Print info
                self.total_ticks += self.iter_ticks
                self.iter_ticks = 0

                # Log history
                if len(self.history) > 0:
                    delta_mem_rss = round(mem_use[0] - self.history[-1][3], 2)
                    delta_mem_vms = round(mem_use[1] - self.history[-1][6], 2)
                    delta_iter_time = round(iter_time - self.history[-1][9], 5)
                    cum_mem_rss = round(mem_use[0] - self.history[0][3], 2)
                    cum_mem_vms = round(mem_use[1] - self.history[0][6], 2)
                    cum_iter_time = round(iter_time - self.history[0][9], 5)
                    self.history.append(
                        [
                            self.pid,
                            self.name,
                            self.total_ticks,
                            round(mem_use[0], 1),
                            delta_mem_rss,
                            cum_mem_rss,
                            round(mem_use[1], 1),
                            delta_mem_vms,
                            cum_mem_vms,
                            iter_time,
                            delta_iter_time,
                            cum_iter_time,
                        ]
                    )
                else:
                    self.history.append(
                        [
                            self.pid,
                            self.name,
                            self.total_ticks,
                            round(mem_use[0], 1),
                            0,
                            0,
                            round(mem_use[1], 1),
                            0,
                            0,
                            iter_time,
                            0,
                            0,
                        ]
                    )
                self.backend.loginfo("\n" + tabulate(self.history, headers=self.headers))
            self.iter_start = time.time()

        # Skip callback if not all inputs with window > 0 have received at least one input.
        if (
            not self.sync or not self.num_ticks > 1
        ):  # Larger than 1 here, because we might have already ticked, but not yet received a skipped input.
            for cname, skip in self.windowed.items():
                if len(kwargs[cname].msgs) == 0:
                    if skip and self.num_ticks == 0:
                        continue
                    else:
                        self.skipped_cbs += 1
                        self.backend.logdebug(f"[{self.name}][{cname}]: skipped_cbs={self.skipped_cbs}")
                        return self.empty_outputs
        output = self.callback(t_n, **kwargs)
        self.num_ticks += 1
        # We skip output type checks for node = "environment":
        #  - The callback sometimes returns None, because .reset() was called instead of a .step().
        if self.name != "environment":
            for cname, _ in self.outputs.items():
                if cname not in output:
                    self.backend.logwarn_once(
                        f"The .callback(...) of `{self.name}` did not return the registered output `{cname}`. "
                        "Downstream nodes, depending on this output for their callback, may deadlock. "
                    )

            for cname, _ in self.targets.items():
                cname_done = cname + "/done"
                assert cname_done in output, (
                    f"The .callback(...) of `{self.name}` did not return an output dict with key"
                    f"`{cname_done}` that communicates the reset status for the registered target."
                )
                assert isinstance(output[cname_done], (bool, np.bool_)), (
                    f"The message type of target {cname_done}` ({type(output[cname_done])}), returned by the .callback(...) of "
                    f"`{self.name}`, should be of type `bool`."
                )
        return output

    def set_delay(self, delay: float, component: str, cname: str) -> None:
        """A method to vary the delay of an input or feedthrough.

        :param delay: A non-negative delay that can be varied at the beginning of an episode (during the reset procedure).
        :param component: Either "inputs" or "feedthroughs".
        :param cname: name of the component.
        """
        assert delay >= 0, "Delay must be non-negative."
        assert cname in getattr(self, component), f"Cannot set the delay. '{cname}' is not one of the {component}."
        getattr(self, component)[cname]["delay"] = delay

    @abc.abstractmethod
    def reset(self, **states: Any) -> None:
        """An abstract method that resets the node to its initial state before the start of an episode.

        This method should be decorated with :func:`eagerx.core.register.states` to register the states.

        :param states: States that were registered (& selected) with the :func:`eagerx.core.register.states` decorator by the subclass.
                       The state messages are sent by the environment and can be used to reset the node at the start of an episode.
                       This can be anything from an estimator's initial state to a hyper-parameter (e.g. delay, control gains).
        """
        pass

    @abc.abstractmethod
    def callback(self, t_n: float, **inputs: Msg) -> Dict[str, Any]:
        """An abstract method that is called at the specified node rate.

        This method should be decorated with:

        - :func:`eagerx.core.register.inputs` to register the inputs.

        - :func:`eagerx.core.register.outputs` to register the outputs.

        :param t_n: Time passed (seconds) since last reset. Increments with 1/:attr:`~eagerx.core.entities.Node.rate`.
        :param inputs: Inputs that were registered (& selected) with the :func:`eagerx.core.register.inputs` decorator by the subclass.
        :return: Dictionary with outputs that were registered (& selected) with the :func:`eagerx.core.register.outputs` decorator by the subclass.
        """
        pass

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        spec = super().pre_make(entity_id, entity_type)
        with spec.config as d:
            d.executable = "python:=eagerx.core.executable_node"
        from eagerx.core.specs import NodeSpec  # noqa: F811

        return NodeSpec(spec.params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)
        entity_id = spec.config.entity_id
        name = spec.config.name

        # Check that there is at least a single input & output defined.
        assert (
            len(spec.config.inputs) > 0 or name == "environment"
        ), f'Node "{name}" does not have any inputs selected. Please select at least one input when making the spec, or check the spec defined for "{entity_id}".'


class ResetNode(Node):
    """Baseclass for nodes that perform a reset routine.

    Use this baseclass to implement reset nodes that will be added to the (agnostic) :class:`~eagerx.core.graph.Graph`.

    Users must call :func:`~eagerx.core.entities.ResetNode.make` to make the reset node subclass' specification.

    .. note:: Subclasses must always have at least one target registered with the :func:`eagerx.core.register.targets` decorator.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.ResetNode.make`

    - :func:`~eagerx.core.entities.ResetNode.initialize`

    - :func:`~eagerx.core.entities.ResetNode.reset`

    - :func:`~eagerx.core.entities.ResetNode.callback`

    - :func:`~eagerx.core.entities.ResetNode.shutdown` (optional)

    Use baseclass :class:`~eagerx.core.entities.EngineNode` instead, for nodes that will be added to
    :class:`~eagerx.core.graph_engine.EngineGraph` when specifying an engine implementation for an :class:`~eagerx.core.entities.Object`.

    Use baseclass :class:`~eagerx.core.entities.Node` instead, for regular nodes that will be added to the
    agnostic :class:`~eagerx.core.graph.Graph`.
    """

    INFO = {
        "make": [],
        "initialize": [],
        "reset": ["states"],
        "callback": ["inputs", "outputs", "targets"],
    }

    def __init__(self, params: Dict, *args, **kwargs):
        super().__init__(*args, params=params, call_init=False, **kwargs)

        # Call initialize with spec
        from eagerx.core.specs import ResetNodeSpec

        self.initialize(ResetNodeSpec(params))

    @abc.abstractmethod
    def reset(self, **states: Any) -> None:
        """An abstract method that resets the node to its initial state before the start of an episode.

        This method should be decorated with :func:`eagerx.core.register.states` to register the states.

        :param states: States that were registered (& selected) with the :func:`eagerx.core.register.states` decorator by the subclass.
                       The state messages are sent by the environment and can be used to reset the node at the start of an episode.
                       This can be anything from an estimator's initial state to a hyper-parameter (e.g. delay, control gains).
        """
        pass

    @abc.abstractmethod
    def callback(self, t_n: float, **inputs_and_targets: Msg) -> Dict[str, Any]:
        """An abstract method that is called at the specified node rate during the environment reset.

        This method should be decorated with:

        - :func:`eagerx.core.register.inputs` to register the inputs.

        - :func:`eagerx.core.register.outputs` to register the outputs.

        - :func:`eagerx.core.register.targets` to register the targets.

        .. note:: This callback is skipped until the user calls :func:`~eagerx.core.env.BaseEnv.reset`.
                  Until then, the messages coming in via the connected *feedthroughs* are fed through as
                  the outputs instead. For every registered output that was registered (& selected) with the
                  :func:`eagerx.core.register.outputs` decorator by the subclass, there must be a connected *feedthrough*.

        :param t_n: Time passed (seconds) since last reset. Increments with 1/:attr:`~eagerx.core.entities.ResetNode.rate`.
        :param inputs_and_targets: Inputs and targets that were registered (& selected) with the :func:`eagerx.core.register.inputs`
                                   and :func:`eagerx.core.register.targets` decorators by the subclass.
        :return: Dictionary with outputs that were registered (& selected) with the :func:`eagerx.core.register.outputs`
                 decorator by the subclass. In addition, the dictionary must contain message of type bool
                 that specifies whether the requested *target* was reached.
        """
        pass

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        spec = super().pre_make(entity_id, entity_type)
        spec._params["targets"] = dict()
        spec._params["feedthroughs"] = dict()
        with spec.config as d:
            d.targets = []

        from eagerx.core.specs import ResetNodeSpec  # noqa: F811

        return ResetNodeSpec(spec.params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)
        entity_id = spec.config.entity_id
        name = spec.config.name

        # Check that there is at least a single target & output was defined.
        assert (
            len(spec.config.outputs) > 0
        ), f'Node "{name}" does not have any outputs selected. Please select at least one output when making the spec, or check the spec defined for "{entity_id}".'
        assert (
            len(spec.config.targets) > 0
        ), f'Node "{name}" does not have any targets selected. Please select at least one target when making the spec, or check the spec defined for "{entity_id}".'

        # Check if all selected targets have an implementation (other components are checked in BaseNode.check_spec())
        # Check that all selected cnames have a corresponding implementation
        for component in ["targets"]:
            for cname in spec.config[component]:
                c = getattr(spec, component)
                msg = (
                    f'"{cname}" was selected as {component[:-1]}'
                    f'but it is not a registered {component[:-1]} of node "{name}". Check the spec of "{entity_id}".'
                )
                assert cname in c, msg


class EngineNode(Node):
    """Baseclass for nodes that are only to be used in combination with a specific engine.

    Users must call :func:`~eagerx.core.entities.EngineNode.make` to make the engine node subclass' specification.

    Use this baseclass to implement nodes that will be added to an :class:`~eagerx.core.graph_engine.EngineGraph`
    when specifying an engine implementation for an :class:`~eagerx.core.entities.Object`.

    These nodes can, optionally, be synchronized with respect to the simulator clock by registering "tick" as an input.

    .. note:: Engine nodes *only* receive a reference to the :attr:`~eagerx.core.entities.EngineNode.simulator` as an
              argument to :func:`~eagerx.core.entities.EngineNode.initialize` when the engine nodes are launched within
              the same process as the engine. See :class:`~eagerx.core.constants.process` for more info.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.EngineNode.make`

    - :func:`~eagerx.core.entities.EngineNode.initialize`

    - :func:`~eagerx.core.entities.EngineNode.reset`

    - :func:`~eagerx.core.entities.EngineNode.callback`

    - :func:`~eagerx.core.entities.EngineNode.shutdown` (optional)

    Use baseclass :class:`~eagerx.core.entities.Node` instead, for nodes that will be added to the
    (agnostic) :class:`~eagerx.core.graph.Graph`.

    Use baseclass :class:`~eagerx.core.entities.ResetNode` instead, for reset routines.
    """

    def __init__(
        self,
        params: Dict,
        *args,
        simulator: Any = None,
        message_broker: RxMessageBroker = None,
        **kwargs: Any,
    ):
        # Call baseclass constructor
        super().__init__(*args, message_broker=message_broker, params=params, call_init=False, **kwargs)

        # Call initialize with spec
        from eagerx.core.specs import NodeSpec

        self.initialize(spec=NodeSpec(params), simulator=simulator)

    @abc.abstractmethod
    def initialize(self, spec: "NodeSpec", simulator: Any) -> None:
        """An abstract method that initializes the node at run-time.

        :param spec: Specification of the engine node.
        :param simulator: A reference to the :attr:`~eagerx.core.entities.Engine.simulator`. The simulator type depends
                          on the engine. Only available if the node was launched inside the engine process.
        """
        pass

    @abc.abstractmethod
    def reset(self, **states: Any) -> None:
        """An abstract method that resets the node to its initial state before the start of an episode.

        This method should be decorated with :func:`eagerx.core.register.states` to register the states.

        .. warning:: Avoid defining states for engine nodes, as you risk making your :class:`~eagerx.core.entities.Object`
                     non-agnostic to the environment. Instead, try to implement object states as an :class:`~eagerx.core.entities.EngineState`
                     of an :class:`~eagerx.core.entities.Object`.

        :param states: States that were registered (& selected) with the :func:`eagerx.core.register.states` decorator by the subclass.
                       The state messages are sent by the environment and can be used to reset the node at the start of an episode.
                       This can be anything from an estimator's initial state to a hyper-parameter (e.g. delay, control gains).
        """
        pass

    @abc.abstractmethod
    def callback(self, t_n: float, **inputs: Msg) -> Dict[str, Any]:
        """An abstract method that is called at the specified node rate.

        This method should be decorated with:

        - :func:`eagerx.core.register.inputs` to register the inputs.

        - :func:`eagerx.core.register.outputs` to register the outputs.

        :param t_n: Time passed (seconds) since last reset. Increments with 1/:attr:`~eagerx.core.entities.EngineNode.rate`.
        :param inputs: Inputs that were registered (& selected) with the :func:`eagerx.core.register.inputs` decorator by the subclass.
        :return: Dictionary with outputs that were registered (& selected) with the :func:`eagerx.core.register.outputs` decorator by the subclass.
        """
        pass

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        spec = super().pre_make(entity_id, entity_type)
        from eagerx.core.specs import NodeSpec  # noqa: F811

        return NodeSpec(spec.params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)


class Engine(BaseNode):
    """Baseclass for engines.

    Use this baseclass to implement an engine that interfaces the simulator.

    Users must call :func:`~eagerx.core.entities.Engine.make` to make the engine subclass' specification.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.Engine.make`

    - :func:`~eagerx.core.entities.Engine.initialize`

    - :func:`~eagerx.core.entities.Engine.add_object`

    - :func:`~eagerx.core.entities.Engine.pre_reset`

    - :func:`~eagerx.core.entities.Engine.reset`

    - :func:`~eagerx.core.entities.Engine.callback`

    - :func:`~eagerx.core.entities.Engine.shutdown` (optional)
    """

    INFO = {
        "make": [],
        "initialize": [],
        "add_object": ["engine_config"],
        "pre_reset": ["states"],
        "reset": ["states"],
        "callback": ["outputs"],
    }

    def __init__(self, sync, real_time_factor, params, target_addresses, node_names, *args, **kwargs):
        """
        Base class constructor.

        :param sync: Boolean flag. Specifies whether we run reactive or asynchronous.
        :param real_time_factor: Sets an upper bound of real_time factor. Wall-clock rate=real_time_factor*rate.
                                 If real_time_factor < 1 the simulation is slower than real time.
        :param target_addresses: Addresses from which the engine should expect "done" msgs.
        :param simulator: Simulator object
        :param node_names: List of node names that are registered in the graph.
        :param kwargs: Arguments that are to be passed down to the node baseclass. See NodeBase for this.
        """
        #: The simulator object. The simulator depends on the engine and should be initialized in the
        #: :func:`~eagerx.core.entities.Engine.initialize` method. Oftentimes, engine nodes require a reference in
        #: :func:`~eagerx.core.entities.EngineNode.callback` and/or :func:`~eagerx.core.entities.EngineNode.reset`
        #: to this simulator object to simulate (e.g. apply an action, extract a sensor measurement).
        #: Engine nodes only have this reference if the node was launched inside the engine process.
        #: See :class:`~eagerx.core.constants.process` for more info.
        self.simulator: Any = dict()
        self.target_addresses = target_addresses
        self.node_names = node_names

        # Check real_time_factor & reactive args
        assert sync or (
            not sync and real_time_factor > 0
        ), "Cannot have a real_time_factor=0 while not reactive. Will result in synchronization issues. Set sync=True or real_time_factor > 0"

        # Initialized nodes
        self.sp_nodes = dict()
        self.launch_nodes = dict()
        self.num_resets = 0
        self.is_initialized = dict()

        # Message counter
        self.dtype_tick = params["outputs"]["tick"]["dtype"]
        self.num_ticks = 0

        # Track memory usage  & speed
        self.total_ticks = 0
        self.py = psutil.Process(os.getpid())
        self.pid = os.getpid()
        self.iter_start = None
        self.iter_ticks = 0
        self.print_iter = 200
        self.history = []
        self.headers = [
            "pid",
            "node",
            "ticks",
            "rss",
            "diff",
            "t0",
            "vms",
            "diff",
            "t0",
            "iter_time",
            "diff",
            "t0",
        ]

        # Call baseclass constructor (which eventually calls .initialize())
        super().__init__(*args, sync=sync, real_time_factor=real_time_factor, params=params, **kwargs)

        # Call initialize with spec
        from eagerx.core.specs import EngineSpec, NodeSpec

        self.initialize(EngineSpec(params))

        #: Parameters for all objects.
        self.objects: dict = params["objects"]

        # Add objects to simulator
        for name, obj in params["objects"].items():
            assert name not in self.simulator, f"There already exists an object called `{name}`."
            self.simulator[name] = dict(name=name)
            self.add_object(**obj["add_object"])

        # Prepare engine states
        for name, obj in params["objects"].items():
            if "engine_states" not in obj:
                continue
            for _cname, i in obj["engine_states"].items():
                if isinstance(i["processor"], dict):
                    from eagerx.core.specs import ProcessorSpec

                    i["processor"] = initialize_processor(ProcessorSpec(i["processor"]))
                if isinstance(i["space"], dict):
                    i["space"] = eagerx.Space.from_dict(i["space"])
                state_cls = load(i["state"]["state_type"])
                i["state"] = state_cls(
                    ns=self.ns, name=i["name"], simulator=self.simulator[name], backend=self.backend, params=i["state"]
                )

        # Initialize nodes
        for node in params["config"]["node_names"]:
            if node in ["environment", "env/supervisor", "engine"]:
                continue
            ns_node = f"{self.ns}/{node}"
            node_params = eagerx.utils.utils.get_param_with_blocking(ns_node, self.backend)
            spec = NodeSpec(node_params)
            if not spec.config.process == eagerx.ENGINE:
                continue
            # If engine node, find out what object name
            if issubclass(load(node_params["node_type"]), EngineNode):
                name = spec.config.name
                name_split = name.split("/")
                assert len(name_split) == 2, f"Cannot infer object name from the engine node's name '{name}'."
                node_args = {"simulator": self.simulator[name_split[0]]}
            else:
                node_args = dict()

            initialize_nodes(
                spec,
                ENGINE,
                self.ns,
                self.message_broker,
                self.is_initialized,
                self.sp_nodes,
                self.launch_nodes,
                node_args=node_args,
            )
        # Wait for all engine nodes to be initialized.
        wait_for_node_initialization(self.is_initialized, self.backend)

    # def register_node(self, node_params):
    #     # If engine node, find out what object name
    #     if issubclass(load(node_params["node_type"]), EngineNode):
    #         name = node_params["config"]["name"]
    #         name_split = name.split("/")
    #         assert len(name_split) == 2, f"Cannot infer object name from the engine node's name '{name}'."
    #         node_args = {"simulator": self.simulator[name_split[0]]}
    #     else:
    #         node_args = dict()
    #
    #     # Initialize nodes
    #     sp_nodes = dict()
    #     launch_nodes = dict()
    #     initialize_nodes(
    #         node_params, ENGINE, self.ns, self.message_broker, self.is_initialized, sp_nodes, launch_nodes, node_args=node_args
    #     )
    #     # todo: move this to end of initialization.
    #     wait_for_node_initialization(self.is_initialized, self.backend)
    #     self.sp_nodes.update(sp_nodes)
    #     self.launch_nodes.update(launch_nodes)
    #     return node_params, sp_nodes, launch_nodes

    def pre_reset_cb(self, **kwargs):
        keys_to_pop = []
        for cname, msg in kwargs.items():
            if msg.info.done:
                keys_to_pop.append(cname)
            else:
                kwargs[cname] = msg.msgs[0]
        [kwargs.pop(key) for key in keys_to_pop]
        return self.pre_reset(**kwargs)

    def reset_cb(self, **kwargs):
        self.num_ticks = 0
        keys_to_pop = []
        for cname, msg in kwargs.items():
            if msg.info.done:
                keys_to_pop.append(cname)
            else:
                kwargs[cname] = msg.msgs[0]
        [kwargs.pop(key) for key in keys_to_pop]
        self.num_resets += 1
        return self.reset(**kwargs)

    def callback_cb(self, node_tick: int, t_n: float, **kwargs: Optional[Msg]):
        self.iter_ticks += 1
        if self.log_memory and self.iter_ticks % self.print_iter == 0:
            if self.iter_start:
                # Time steps
                iter_stop = time.time()
                if self.iter_ticks > 0:
                    iter_time = float((iter_stop - self.iter_start) / self.iter_ticks)
                else:
                    iter_time = float(iter_stop - self.iter_start)

                # Memory usage request
                mem_use = (np.array(self.py.memory_info()[0:2]) / 2.0**30) * 1000  # memory use in MB...I think

                # Print info
                self.total_ticks += self.iter_ticks
                self.iter_ticks = 0

                # Log history
                if len(self.history) > 0:
                    delta_mem_rss = round(mem_use[0] - self.history[-1][3], 2)
                    delta_mem_vms = round(mem_use[1] - self.history[-1][6], 2)
                    delta_iter_time = round(iter_time - self.history[-1][9], 5)
                    cum_mem_rss = round(mem_use[0] - self.history[0][3], 2)
                    cum_mem_vms = round(mem_use[1] - self.history[0][6], 2)
                    cum_iter_time = round(iter_time - self.history[0][9], 5)
                    self.history.append(
                        [
                            self.pid,
                            self.name,
                            self.total_ticks,
                            round(mem_use[0], 1),
                            delta_mem_rss,
                            cum_mem_rss,
                            round(mem_use[1], 1),
                            delta_mem_vms,
                            cum_mem_vms,
                            iter_time,
                            delta_iter_time,
                            cum_iter_time,
                        ]
                    )
                else:
                    self.history.append(
                        [
                            self.pid,
                            self.name,
                            self.total_ticks,
                            round(mem_use[0], 1),
                            0,
                            0,
                            round(mem_use[1], 1),
                            0,
                            0,
                            iter_time,
                            0,
                            0,
                        ]
                    )
                self.backend.get_log_fn(self.log_level)("\n" + tabulate(self.history, headers=self.headers))
                # backend.loginfo("\n" + tabulate(self.history, headers=self.headers))
            self.iter_start = time.time()
        # Only apply the callback after all pipelines have been initialized
        # Only then, the initial state has been set.
        if self.num_resets >= 1 and node_tick > 0:
            self.callback(t_n)
        # Fill output msg with number of node ticks
        self.num_ticks += 1
        return dict(tick=np.array(node_tick, dtype=self.dtype_tick))

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        spec = super().pre_make(entity_id, entity_type)

        # Add objects key
        from eagerx.core.specs import EngineSpec  # noqa: F811

        params = spec.params
        params["objects"] = dict()
        spec = EngineSpec(params)

        # Set default engine params
        with spec.config as d:
            d.name = "engine"
            d.sync = True
            d.real_time_factor = 0
            d.simulate_delays = True
            d.executable = "python:=eagerx.core.executable_engine"

        # Add tick as output
        space = eagerx.Space(shape=(), dtype="int64")
        spec.add_output("tick", space=space)
        spec.config.outputs.append("tick")
        return EngineSpec(spec.params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)

    @abc.abstractmethod
    def add_object(
        self,
        name: str,
        *args: Union[bool, int, float, str, List, Dict],
        **kwargs: Union[bool, int, float, str, List, Dict],
    ) -> None:
        """Adds an object to the simulator that is interfaced by the engine.

        :param name: The name of the :class:`~eagerx.core.entities.Object` that is to be added.
        :param args: The engine-specific parameters that are required to add the :class:`~eagerx.core.entities.Object`.
        :param kwargs: The engine-specific parameters that are optional to add the :class:`~eagerx.core.entities.Object`.
        """
        pass

    @abc.abstractmethod
    def pre_reset(self, **states: Any) -> None:
        """
        An abstract method that resets the engine to its initial state before the start of an episode.

        .. note:: This method is called **before** every :class:`~eagerx.core.entities.EngineNode` and
                  :class:`~eagerx.core.entities.EngineState` has performed its reset,
                  but after all reset routines, implemented with :class:`~eagerx.core.entities.ResetNode`,
                  have reached their target.

                  - Can be useful for performing some preliminary actions on the simulator such as pausing before resetting
                    every :class:`~eagerx.core.entities.EngineNode` and :class:`~eagerx.core.entities.EngineState`.

                  - Reset the simulator state so that this state can be used in the reset of every
                    :class:`~eagerx.core.entities.EngineNode` and :class:`~eagerx.core.entities.EngineState`.

        :param states: States that were registered (& selected) with the :func:`eagerx.core.register.states` decorator by the subclass.
                       The state messages are sent by the environment and can be used to reset the engine at the start of an episode.
                       This can be anything, such as the dynamical properties of the simulator (e.g. friction coefficients).
        """
        pass

    @abc.abstractmethod
    def reset(self, **states: Any) -> None:
        """An abstract method that resets the engine to its initial state before the start of an episode.

        This method should be decorated with :func:`eagerx.core.register.states` to register the states.

        .. note:: This method is called **after** every :class:`~eagerx.core.entities.EngineNode` and
                  :class:`~eagerx.core.entities.EngineState` has finished its reset.

                  - Can be useful for performing some final actions on the simulator such as unpausing after every
                    :class:`~eagerx.core.entities.EngineNode` and :class:`~eagerx.core.entities.EngineState` have reset.

        :param states: States that were registered (& selected) with the :func:`eagerx.core.register.states` decorator by the subclass.
                       The state messages are sent by the environment and can be used to reset the engine at the start of an episode.
                       This can be anything, such as the dynamical properties of the simulator (e.g. friction coefficients).
        """
        pass

    @abc.abstractmethod
    def callback(self, t_n: float) -> None:
        """
        The engine callback that is performed at the specified rate.

        This callback is steps the simulator by 1/:attr:`~eagerx.core.entities.Engine.rate`.

        .. note:: The engine does not have any outputs.
                  If you wish to broadcast other output messages based on properties of the simulator,
                  a separate :class:`~eagerx.core.entities.EngineNode` should be created.

        :param t_n: Time passed (seconds) since last reset. Increments with 1/:attr:`~eagerx.core.entities.Engine.rate`.
        """
        pass


class Object(Entity):
    """Baseclass for objects.

    Use this baseclass to implement objets that consist of sensors, actuators, and/or engine states.

    Users must call :func:`~eagerx.core.entities.Object.make` to make the object subclass' specification.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.Object.make`

    For every supported :class:`~eagerx.core.entities.Engine`, an implementation method must be added.
    This method must have the same signature as :func:`~eagerx.core.entities.Object.example_engine`:

    - :func:`~eagerx.core.entities.Object.pybullet` (example)

    - :func:`~eagerx.core.entities.Object.ode_engine` (example)

    - ...
    """

    INFO = {
        "make": ["sensors", "actuators", "engine_states"],
    }

    @classmethod
    @abc.abstractmethod
    def make(cls, *args: Any, **kwargs: Any):
        """An abstract method that makes the specification (also referred to as `spec`) of this object.

        See :class:`~eagerx.core.specs.ObjectSpec` how sensor/actuator/engine state parameters can be set.

        This method should be decorated with:

        - :func:`eagerx.core.register.sensors` to register sensors.

        - :func:`eagerx.core.register.actuators` to register actuators.

        - :func:`eagerx.core.register.engine_states` to register engine states.

        :param args: Arguments to the subclass' make function.
        :param kwargs: Optional Arguments to the subclass' make function.
        :return: A (mutable) spec that can be used to build and subsequently initialize the entity (e.g. node, object, ...).
        """
        pass

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        spec = super().pre_make(entity_id, entity_type)
        params = spec.params
        params["config"] = dict(
            name=None,
            sensors=[],
            actuators=[],
            states=[],
            entity_id=params.pop("entity_id"),
        )
        params["sensors"] = dict()
        params["actuators"] = dict()
        params["states"] = dict()
        from eagerx.core.specs import ObjectSpec  # noqa: F811

        return ObjectSpec(params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)
        entity_id = spec.config.entity_id
        name = spec.config.name

        # Check that all components have a space
        for component in ["states"]:
            for cname, params in getattr(spec, component).items():
                msg = (
                    f'"{cname}" was defined as {component[:-1]} in "{name}", but its space ({type(params.space)}) '
                    f'is not a valid space. Check the spec of "{entity_id}".'
                )
                assert params.space is not None, msg

    def example_engine(self, spec: "ObjectSpec", graph: "EngineGraph") -> None:
        """An example of an engine-specific implementation of an object's registered sensors, actuators, and/or states.

        See :attr:`~eagerx.core.specs.ObjectSpec.engine` how engine specific parameters can be set/get.

        This method must be decorated with :func:`eagerx.core.register.engine` to register
        the engine implementation of the object.

        .. note:: This is an example method for documentation purposes only.

        :param spec: A (mutable) specification.
        :param graph: A graph containing the object's registered (disconnected) sensors & actuators.
                      Users should add nodes that inherit from :class:`~eagerx.core.entities.EngineNode`, and connect
                      them to the sensors & actuators. As such, the engine nodes become the *engine-specific implementation*
                      of the agnostic sensors & actuator definition.
        """
        raise NotImplementedError("This is a mock engine implementation for documentation purposes.")


class Processor(Entity):
    """Baseclass for processors.

    Use this baseclass to implement processor that preprocess an input/output message.

    This baseclass only supports one-way processing.

    Users must call :func:`~eagerx.core.entities.Processor.make` to make the subclass' specification.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.Processor.make`

    - :func:`~eagerx.core.entities.Processor.initialize`

    - :func:`~eagerx.core.entities.Processor.convert`

    """

    INFO = {
        "make": [],
        "initialize": [],
        "convert": [],
    }

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def initialize(self, spec: "ProcessorSpec") -> None:
        """An abstract method to initialize the processor.

        :param spec: Specification of the processor.
        """
        pass

    @abc.abstractmethod
    def convert(self, msg: Any) -> Any:
        """An abstract method to preprocess messages.

        :param msg: Raw message.
        :return: Preprocessed message.
        """
        pass

    @classmethod
    def pre_make(cls, entity_id: str, entity_type: "Entity"):
        spec = super().pre_make(entity_id, entity_type)
        params = spec.params
        params["processor_type"] = params.pop("entity_type")
        params.pop("entity_id")
        from eagerx.core.specs import ProcessorSpec  # noqa: F811

        return ProcessorSpec(params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)


class EngineState(Entity):
    """Baseclass for engine states.

    Use this baseclass to implement engine states for an :class:`~eagerx.core.entities.Object`.

    Users must call :func:`~eagerx.core.entities.EngineState.make` to make the subclass' specification.

    Subclasses must implement the following methods:

    - :func:`~eagerx.core.entities.EngineState.make`

    - :func:`~eagerx.core.entities.EngineState.initialize`

    - :func:`~eagerx.core.entities.EngineState.reset`

    """

    INFO = {
        "make": [],
        "initialize": [],
        "reset": [],
    }

    def __init__(
        self,
        ns,
        name,
        simulator,
        backend,
        params,
    ):
        #: Namespace of the environment. Can be set with the `name` argument to :class:`~eagerx.core.env.BaseEnv`.
        self.ns = ns
        #: Name of the state.
        self.name = name
        self.color = "grey"
        #: Responsible for all I/O communication within this process.
        self.backend = backend
        from eagerx.core.specs import EngineStateSpec

        self.initialize(EngineStateSpec(params), simulator)

    @classmethod
    def pre_make(cls, entity_id, entity_type):
        spec = super().pre_make(entity_id, entity_type)
        params = spec.params
        params["state_type"] = params.pop("entity_type")
        params.pop("entity_id")
        from eagerx.core.specs import EngineStateSpec  # noqa: F811

        return EngineStateSpec(params)

    @classmethod
    def check_spec(cls, spec):
        super().check_spec(spec)
        return

    @abc.abstractmethod
    def initialize(self, spec: "EngineStateSpec", simulator: Any) -> None:
        """An abstract method to initialize the engine state.

        :param spec: The engine state specification.
        :param simulator: A reference to the engine's simulator.
        """
        pass

    @abc.abstractmethod
    def reset(self, state: Any) -> None:
        """An abstract method to reset the engine state of an :class:`~eagerx.core.entities.Object`.

        :param state: The desired state that the user can specify before calling :func:`~eagerx.core.env.BaseEnv.reset`.
        """
        pass



import logging
import inspect
import pickle

# EAGERx
from eagerx.core.constants import (
    DEBUG,
    INFO,
    ERROR,
    WARN,
    FATAL,
)


def get_log_level():
    return logging.getLogger("eagerx").getEffectiveLevel()


def set_log_level(log_level):
    logger = logging.getLogger("eagerx")
    logger.setLevel(log_level)


def _log(msg, level, effective_level=None):
    if level >= get_log_level():
        print(msg)


def logdebug(msg):
    return _log(msg, DEBUG)


def loginfo(msg):
    return _log(msg, INFO)


def logwarn(msg):
    return _log(msg, WARN)


def logerr(msg):
    return _log(msg, ERROR)


def logfatal(msg):
    return _log(msg, FATAL)


def logsilent(msg, *args, **kwargs):
    pass


def frame_to_caller_id(frame):
    caller_id = (
        inspect.getabsfile(frame),
        frame.f_lineno,
        frame.f_lasti,
    )
    return pickle.dumps(caller_id)


class LoggingOnce(object):

    called_caller_ids = set()

    def __call__(self, caller_id):
        if caller_id not in self.called_caller_ids:
            self.called_caller_ids.add(caller_id)
            return True
        return False



from eagerx.core.constants import process  # noqa: F401
from eagerx.core.entities import (  # noqa: F401
    Object,
    Engine,
    Node,
    ResetNode,
    Processor,
    EngineState,
    EngineNode,
)



"""
Adapted from the roslaunch.substitution_args package.
"""

import yaml
import math
import ast
import os
from io import StringIO
from typing import Union, Dict, Optional, List


class SubstitutionException(Exception):
    """
    Base class for exceptions in substitution_args routines
    """

    pass


class ArgException(SubstitutionException):
    """
    Exception for missing $(arg) values
    """

    pass


def substitute_args(
    param: Union[str, Dict],
    context: Optional[Dict] = None,
    only: Optional[List[str]] = None,
):
    """Substitute arguments based on the context dictionairy (and possibly sourced packages).
    Follows the xacro substition convention of ROS, with a 'config' and 'ns' command added.
    :param param: dict or string we wish to perform substitutions on.
    :param context: dict[command][context] with replacement values .
    :param only: List of possible commands. If only is not provided, all commands are executed.
                 Options are ['env', 'optenv', 'dirname', 'anon', 'arg', 'ns', 'config'].
    :return: Substituted param file
    """
    # substitute string
    if isinstance(param, str):
        param = resolve_args(param, context, only=only)
        return param

    # For every key in the dictionary (not performing copy.deepcopy!)
    if isinstance(param, dict):
        for key in param:
            # If the value is of type `(Ordered)dict`, then recurse with the value
            if isinstance(param[key], (dict, list)):
                substitute_args(param[key], context, only=only)
            # Otherwise, add the element to the result
            elif isinstance(param[key], str):
                param[key] = resolve_args(param[key], context, only=only)
    elif isinstance(param, list):
        for idx, i in enumerate(param):
            if isinstance(i, (dict, list)):
                substitute_args(i, context, only=only)
            elif isinstance(i, str):
                param[idx] = resolve_args(i, context, only=only)
    return param


def resolve_args(arg_str, context=None, resolve_anon=True, filename=None, only=None):
    """
    Resolves substitution args (see wiki spec U{http://ros.org/wiki/roslaunch}).

    @param arg_str: string to resolve zero or more substitution args
        in. arg_str may be None, in which case resolve_args will
        return None
    @type  arg_str: str
    @param context dict: (optional) dictionary for storing results of
        the 'anon' and 'arg' substitution args. multiple calls to
        resolve_args should use the same context so that 'anon'
        substitions resolve consistently. If no context is provided, a
        new one will be created for each call. Values for the 'arg'
        context should be stored as a dictionary in the 'arg' key.
    @type  context: dict
    @param resolve_anon bool: If True (default), will resolve $(anon
        foo). If false, will leave these args as-is.
    @type  resolve_anon: bool

    @return str: arg_str with substitution args resolved
    @rtype:  str
    @raise SubstitutionException: if there is an error resolving substitution args
    """
    if context is None:
        context = {}
    if not arg_str:
        return arg_str
    # special handling of $(eval ...)
    if arg_str.startswith("$(eval ") and arg_str.endswith(")"):
        return _eval(arg_str[7:-1], context)
    # first resolve variables like 'env' and 'arg'
    commands = {
        "env": _env,
        "optenv": _optenv,
        "dirname": _dirname,
        "arg": _arg,
        "ns": _ns,
        "config": _config,
    }
    if only is not None:
        exec_commands = {}
        for c in only:
            if c in commands:
                exec_commands[c] = commands[c]
    else:
        exec_commands = commands
    resolved = _resolve_args(arg_str, context, resolve_anon, exec_commands)
    # then resolve 'find' as it requires the subsequent path to be expanded already
    commands = {
        "find": _find,
    }
    if only is not None:
        exec_commands = {}
        for c in only:
            if c in commands:
                exec_commands[c] = commands[c]
    else:
        exec_commands = commands
    resolved = _resolve_args(resolved, context, resolve_anon, exec_commands)
    return resolved


def _find(resolved, a, args, context):
    try:
        import roslaunch
    except ImportError as e:
        msg = f"You can only use $(find ...) if ROS is sourced: {e}"
        raise ImportError(msg)
    return roslaunch.substitution_args._find(resolved, a, args, context)


class _DictWrapper(object):
    def __init__(self, args, functions):
        self._args = args
        self._functions = functions

    def __getitem__(self, key):
        try:
            return self._functions[key]
        except KeyError:
            return convert_value(self._args[key], "auto")


def convert_value(value, type_):
    """
    Convert a value from a string representation into the specified
    type

    @param value: string representation of value
    @type  value: str
    @param type_: int, double, string, bool, or auto
    @type  type_: str
    @raise ValueError: if parameters are invalid
    """
    type_ = type_.lower()
    # currently don't support XML-RPC date, dateTime, maps, or list
    # types
    if type_ == "auto":
        # attempt numeric conversion
        try:
            if "." in value:
                return float(value)
            else:
                return int(value)
        except ValueError:
            pass
        # bool
        lval = value.lower()
        if lval == "true" or lval == "false":
            return convert_value(value, "bool")
        # string
        return value
    elif type_ == "str" or type_ == "string":
        return value
    elif type_ == "int":
        return int(value)
    elif type_ == "double":
        return float(value)
    elif type_ == "bool" or type_ == "boolean":
        value = value.lower().strip()
        if value == "true" or value == "1":
            return True
        elif value == "false" or value == "0":
            return False
        raise ValueError("%s is not a '%s' type" % (value, type_))
    elif type_ == "yaml":
        try:
            return yaml.safe_load(value)
        except yaml.parser.ParserError as e:
            raise ValueError(e)
    else:
        raise ValueError("Unknown type '%s'" % type_)


def _eval(s, context):
    if "anon" not in context:
        context["anon"] = {}
    if "arg" not in context:
        context["arg"] = {}

    # inject arg context
    def _eval_arg_context(name):
        return convert_value(_eval_arg(name, args=context["arg"]), "auto")

    # inject dirname context
    def _eval_dirname_context():
        return _eval_dirname(context["filename"])

    functions = {"arg": _eval_arg_context, "dirname": _eval_dirname_context}
    functions.update(_eval_dict)

    # ignore values containing double underscores (for safety)
    # http://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html
    if s.find("__") >= 0:
        raise SubstitutionException("$(eval ...) may not contain double underscore expressions")
    return str(eval(s, {}, _DictWrapper(context["arg"], functions)))


def _eval_env(name):
    try:
        return os.environ[name]
    except KeyError as e:
        raise SubstitutionException("environment variable %s is not set" % str(e))


def _env(resolved, a, args, context):
    """
    process $(env) arg
    @return: updated resolved argument
    @rtype: str
    @raise SubstitutionException: if arg invalidly specified
    """
    if len(args) != 1:
        raise SubstitutionException("$(env var) command only accepts one argument [%s]" % a)
    return resolved.replace("$(%s)" % a, _eval_env(args[0]))


def _eval_dirname(filename):
    if not filename:
        raise SubstitutionException("Cannot substitute $(dirname), no file/directory information available.")
    return os.path.abspath(os.path.dirname(filename))


def _dirname(resolved, a, args, context):
    """
    process $(dirname)
    @return: updated resolved argument
    @rtype: str
    @raise SubstitutionException: if no information about the current launch file is available, for example
           if XML was passed via stdin, or this is a remote launch.
    """
    return resolved.replace("$(%s)" % a, _eval_dirname(context.get("filename", None)))


def _eval_optenv(name, default=""):
    if name in os.environ:
        return os.environ[name]
    return default


def _optenv(resolved, a, args, context):
    """
    process $(optenv) arg
    @return: updated resolved argument
    @rtype: str
    @raise SubstitutionException: if arg invalidly specified
    """
    if len(args) == 0:
        raise SubstitutionException("$(optenv var) must specify an environment variable [%s]" % a)
    return resolved.replace("$(%s)" % a, _eval_optenv(args[0], default=" ".join(args[1:])))


def _eval_arg(name, args):
    try:
        return args[name]
    except KeyError:
        raise ArgException(name)


def _arg(resolved, a, args, context):
    """
    process $(arg) arg

    :returns: updated resolved argument, ``str``
    :raises: :exc:`ArgException` If arg invalidly specified
    """
    if len(args) == 0:
        raise SubstitutionException("$(arg var) must specify a variable name [%s]" % (a))
    elif len(args) > 1:
        raise SubstitutionException("$(arg var) may only specify one arg [%s]" % (a))

    if "arg" not in context:
        context["arg"] = {}
    return resolved.replace("$(%s)" % a, _eval_arg(name=args[0], args=context["arg"]))


def _resolve_args(arg_str, context, resolve_anon, commands):
    ros_valid = ["find", "env", "optenv", "dirname", "anon", "arg"]
    valid = ros_valid + ["ns", "config"]
    resolved = arg_str
    if isinstance(arg_str, (str, list)):
        for a in _collect_args(arg_str):
            splits = [s for s in a.split(" ") if s]
            if not splits[0] in valid:
                raise SubstitutionException("Unknown substitution command [%s]. Valid commands are %s" % (a, valid))
            command = splits[0]
            args = splits[1:]
            if command in commands:
                resolved = commands[command](resolved, a, args, context)
    return resolved


def _eval_config(name, args):
    try:
        return args[name]
    except KeyError:
        raise ArgException(name)


_eval_ns = _eval_config


def tryeval(val):
    try:
        val = ast.literal_eval(val)
    except (ValueError, SyntaxError):
        pass
    return val


def _config(resolved, a, args, context):
    """
    process $(config) arg

    :returns: updated resolved argument, ``str``
    :raises: :exc:`ArgException` If arg invalidly specified
    """
    if len(args) == 0:
        raise SubstitutionException("$(config var) must specify a variable name [%s]" % (a))
    elif len(args) > 1:
        raise SubstitutionException("$(config var) may only specify one arg [%s]" % (a))

    if "config" not in context:
        context["config"] = {}
    try:
        return tryeval(resolved.replace("$(%s)" % a, str(_eval_config(name=args[0], args=context["config"]))))
    except ArgException as e:  # ArgException:
        if isinstance(e, ArgException):
            return resolved
        elif isinstance(e, TypeError):
            raise
        else:
            raise


def _ns(resolved, a, args, context):
    """
    process $(ns) arg

    :returns: updated resolved argument, ``str``
    :raises: :exc:`ArgException` If arg invalidly specified
    """
    if len(args) == 0:
        raise SubstitutionException("$(ns var) must specify a variable name [%s]" % (a))
    elif len(args) > 1:
        raise SubstitutionException("$(ns var) may only specify one arg [%s]" % (a))

    if "ns" not in context:
        context["ns"] = {}
    try:
        return tryeval(resolved.replace("$(%s)" % a, str(_eval_ns(name=args[0], args=context["ns"]))))
    except ArgException as e:  # ArgException:
        if isinstance(e, ArgException):
            return resolved
        elif isinstance(e, TypeError):
            raise
        else:
            raise


_OUT = 0
_DOLLAR = 1
_LP = 2
_IN = 3


def _collect_args(arg_str):
    """
    State-machine parser for resolve_args. Substitution args are of the form:
    $(find package_name)/scripts/foo.py $(export some/attribute blar) non-relevant stuff

    @param arg_str: argument string to parse args from
    @type  arg_str: str
    @raise SubstitutionException: if args are invalidly specified
    @return: list of arguments
    @rtype: [str]
    """
    buff = StringIO()
    args = []
    state = _OUT
    for c in arg_str:
        # No escapes supported
        if c == "$":
            if state == _OUT:
                state = _DOLLAR
            elif state == _DOLLAR:
                pass
            else:
                raise SubstitutionException("Dollar signs '$' cannot be inside of substitution args [%s]" % arg_str)
        elif c == "(":
            if state == _DOLLAR:
                state = _LP
            elif state != _OUT:
                raise SubstitutionException("Invalid left parenthesis '(' in substitution args [%s]" % arg_str)
        elif c == ")":
            if state == _IN:
                # save contents of collected buffer
                args.append(buff.getvalue())
                buff.truncate(0)
                buff.seek(0)
                state = _OUT
            else:
                state = _OUT
        elif state == _DOLLAR:
            # left paren must immediately follow dollar sign to enter _IN state
            state = _OUT
        elif state == _LP:
            state = _IN

        if state == _IN:
            buff.write(c)
    return args


# Create a dictionary of global symbols that will be available in the eval
# context.  We disable all the builtins, then add back True and False, and also
# add true and false for convenience (because we accept those lower-case strings
# as boolean values in XML).
_eval_dict = {
    "true": True,
    "false": False,
    "True": True,
    "False": False,
    "__builtins__": {k: __builtins__[k] for k in ["list", "dict", "map", "str", "float", "int"]},
    "env": _eval_env,
    "optenv": _eval_optenv,
}
# also define all math symbols and functions
_eval_dict.update(math.__dict__)
